"""Multi-view LVM with informed structured sparsity."""

import logging
import logging.config

from pathlib import Path

from muvi.core import *  # noqa: F403
from muvi.tools import external as ext
from muvi.tools import feature_sets as fs
from muvi.tools import load
from muvi.tools import plotting as pl
from muvi.tools import save
from muvi.tools import utils as tl


logging.config.fileConfig(
    Path(__file__).resolve().parent / "log.conf",
    disable_existing_loggers=False,
)
logging.getLogger().setLevel(logging.INFO)

__version__ = "0.1.1"

__all__ = ["fs", "pl", "tl", "ext", "save", "load"]



import logging

from pathlib import Path

import h5py
import numpy as np

from muvi.tools.utils import variance_explained


logger = logging.getLogger(__name__)


def save_as_hdf5(
    model,
    path,
    save_metadata=True,
):
    if not model._trained:
        raise ValueError(
            "Cannot save an untrained model, call `fit` first to train a MuVI model."
        )

    path = Path(path)
    if path.exists():
        logger.warning(f"`{path}` already exists, overwriting.")

    default_group_name = "group_0"
    logger.info(
        f"Setting default group name to `{default_group_name}` for single group data."
    )
    logger.info("Computing variance explained.")
    r2_view, r2_factor, _ = variance_explained(
        model, subsample=0 if model.n_samples < 5000 else 1000
    )

    f = h5py.File(path, "w")

    # data
    f_data = f.create_group("data")

    for vn in model.view_names:
        f_data.create_group(vn).create_dataset(
            default_group_name, data=model.get_observations()[vn]
        )

    # samples_metadata
    if save_metadata:
        metadata = model._cache.factor_adata.obs
        if not metadata.empty:
            f_metadata = f.create_group("samples_metadata/group_0")
            f_metadata.create_dataset("sample", data=model.sample_names.tolist())
            for col in metadata.columns:
                f_metadata.create_dataset(col, data=metadata[col].to_numpy())

    # views
    f.create_group("views").create_dataset("views", data=model.view_names.tolist())

    # groups
    f.create_group("groups").create_dataset("groups", data=[default_group_name])

    # features
    f_features = f.create_group("features")

    for vn in model.view_names:
        f_features.create_dataset(vn, data=model.feature_names[vn].tolist())

    # samples
    f.create_group("samples").create_dataset(
        default_group_name, data=model.sample_names.tolist()
    )

    # sorted factors
    r2_order = r2_factor.sum(1).argsort().to_numpy()[::-1]
    f.create_group("factors").create_dataset(
        default_group_name, data=model.factor_names[r2_order].tolist()
    )

    # model_options

    model_options = {
        "ard_weights": [True],
        "spikeslab_weights": [True],
        "ard_factors": [False],
        "spikeslab_factors": [False],
        "likelihoods": [
            {"normal": "gaussian"}.get(model.likelihoods[vn], model.likelihoods[vn])
            for vn in model.view_names
        ],
    }

    f_model_options = f.create_group("model_options")
    for name, value in model_options.items():
        f_model_options.create_dataset(name, data=value)

    # training_opts
    training_opts = [
        # maxiter
        model._training_log["n_epochs"],
        # freqELBO
        1,
        # start_elbo
        1,
        # gpu_mode
        int(str(model.device) != "cpu"),
        # stochastic
        int(model._training_log["batch_size"] < model.n_samples),
        # seed
        model._training_log["seed"],
    ]

    f.create_dataset("training_opts", data=training_opts)

    # expectations
    f_w = f.create_group("expectations/W")
    f_z = f.create_group("expectations/Z")

    for vn in model.view_names:
        f_w.create_dataset(vn, data=model.get_factor_loadings()[vn][r2_order, :])
    f_z.create_dataset(
        default_group_name, data=model.get_factor_scores()[:, r2_order].T
    )

    # intercepts
    f_intercepts = f.create_group("intercepts")

    for vn in model.view_names:
        f_intercepts.create_group(vn).create_dataset(
            default_group_name, data=np.zeros(model.n_features[vn])
        )

    # training stats
    n_iter = model._training_log["n_iter"]
    training_stats = {
        "elbo": model._training_log["history"],
        "number_factors": model.n_factors * np.ones(n_iter),
        "time": np.zeros(n_iter),
    }

    f_training_stats = f.create_group("training_stats")
    for name, value in training_stats.items():
        f_training_stats.create_dataset(name, data=value)

    # variance_explained
    f.create_group("variance_explained/r2_total").create_dataset(
        default_group_name, data=[r2_view[vn] for vn in model.view_names]
    )
    f.create_group("variance_explained/r2_per_factor").create_dataset(
        default_group_name, data=r2_factor.iloc[r2_order, :].to_numpy().T
    )

    return f.close()



"""Collection of plotting functions."""
import logging

from typing import Optional
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns

from sklearn.metrics import confusion_matrix as sklearn_cf

import muvi

from muvi.core.index import _normalize_index


sns.set_theme()
sns.set_style("whitegrid")

logger = logging.getLogger(__name__)


HEATMAP = "heatmap"
MATRIXPLOT = "matrixplot"
DOTPLOT = "dotplot"
TRACKSPLOT = "tracksplot"
VIOLIN = "violin"
STACKED_VIOLIN = "stacked_violin"
PL_TYPES = [HEATMAP, MATRIXPLOT, DOTPLOT, TRACKSPLOT, VIOLIN, STACKED_VIOLIN]


def savefig_or_show(
    writekey: str,
    show: Optional[bool] = None,
    dpi: Optional[int] = None,
    ext: Optional[str] = None,
    save: Union[bool, str, None] = None,
):
    return sc.pl._utils.savefig_or_show(writekey, show, dpi, ext, save)


def _subset_df(data, groupby, groups, include_rest=True):
    if groups is None:
        return data

    if include_rest:
        data[groupby] = data[groupby].cat.add_categories(include_rest)
        data.loc[~data[groupby].isin(groups), groupby] = include_rest
        groups.append(include_rest)
    data = data.loc[data[groupby].isin(groups), :].copy()
    data[groupby] = data[groupby].cat.remove_unused_categories()

    if data.empty:
        raise ValueError("Empty data, check whether the provided `groups` are correct.")

    return data


def _setup_legend(
    g, bbox_to_anchor=(1, 0.5), loc="center left", frameon=False, remove_last=False
):
    kwargs = {"bbox_to_anchor": bbox_to_anchor, "loc": loc, "frameon": frameon}

    if remove_last:
        handles, labels = g.get_legend_handles_labels()
        kwargs["handles"] = handles[:-1]
        kwargs["labels"] = labels[:-1]
    g.legend(**kwargs)

    return g


def _get_color_dict(factor_adata, groupby, include_rest=True):
    uns_colors_key = f"{groupby}_colors"
    if uns_colors_key not in factor_adata.uns:
        return None
    color_dict = dict(
        zip(
            factor_adata.obs[groupby].astype("category").cat.categories,
            factor_adata.uns[uns_colors_key],
        )
    )
    if include_rest:
        color_dict[include_rest] = "#D3D3D3"
    return color_dict


def _get_model_cache(model):
    cache = model._cache
    if cache is None:
        raise ValueError("Model cache not initialised, execute muvi.tl.* first.")
    return cache


def _lines(ax, positions, ymin, ymax, horizontal=False, **kwargs):
    if horizontal:
        ax.hlines(positions, ymin, ymax, **kwargs)
    else:
        ax.vlines(positions, ymin, ymax, **kwargs)
    return ax


def lined_heatmap(data, figsize=None, hlines=None, vlines=None, **kwargs):
    """Plot heatmap with horizontal or vertical lines."""
    if figsize is None:
        figsize = (20, 2)
    fig, g = plt.subplots(figsize=figsize)
    g = sns.heatmap(data, ax=g, **kwargs)
    if hlines is not None:
        _lines(
            g,
            hlines,
            *sorted(g.get_xlim()),
            horizontal=True,
            lw=1.0,
            linestyles="dashed",
        )
    if vlines is not None:
        _lines(
            g,
            vlines,
            *sorted(g.get_ylim()),
            horizontal=False,
            lw=1.0,
            linestyles="dashed",
        )
    return g


def variance_explained(
    model,
    top=50,
    show: Optional[bool] = None,
    save: Union[bool, str, None] = None,
    **kwargs,
):
    """Heatmap of variance explained, see `muvi.tl.variance_explained`."""

    model_cache = _get_model_cache(model)
    r2_fac = model_cache.factor_metadata
    r2_cov = model_cache.cov_metadata
    if r2_fac is None and r2_cov is None:
        raise ValueError(
            "No scores found in model cache, rerun `muvi.tl.variance_explained`."
        )

    r2_columns = [f"{model_cache.METRIC_R2}_{vn}" for vn in model.view_names]
    if r2_fac is not None:
        r2_fac = r2_fac.loc[:, r2_columns].T
    if r2_cov is not None:
        r2_cov = r2_cov.loc[:, r2_columns].T

    if top > 0:
        r2_argsort = r2_fac.sum().argsort()[::-1]
        r2_fac = r2_fac.iloc[:, r2_argsort]
    if top > 1:
        r2_fac = r2_fac.iloc[:, :top]

    r2_joint = r2_fac
    if r2_cov is not None:
        r2_joint = pd.concat([r2_fac, r2_cov], axis=1)

    if r2_joint.isna().all(None):
        raise ValueError(
            "No scores found in model cache, rerun `muvi.tl.variance_explained`."
        )

    r2_joint.index = r2_joint.index.str[len(model_cache.METRIC_R2) + 1 :]

    figsize = (max(20, r2_joint.shape[1] // 5), r2_joint.shape[0])
    vlines = r2_fac.shape[1]
    if r2_cov is None or r2_cov.empty:
        vlines = None

    kwargs["vmin"] = 0

    g = lined_heatmap(r2_joint, figsize=figsize, vlines=vlines, **kwargs)
    savefig_or_show("variance_explained", show=show, save=save)
    if not show:
        return g


def variance_explained_grouped(
    model,
    factor_idx,
    view_idx="all",
    groups=None,
    kind="bar",
    stacked=True,
    show: Optional[bool] = None,
    save: Union[bool, str, None] = None,
    **kwargs,
):
    model_cache = _get_model_cache(model)
    if model_cache.UNS_GROUPED_R2 not in model_cache.uns:
        raise ValueError(
            f"`{model_cache.UNS_GROUPED_R2}` not found in model cache, "
            "rerun `muvi.tl.variance_explained_grouped`."
        )

    view_idx = _normalize_index(view_idx, model.view_names, as_idx=False)
    factor_idx = _normalize_index(factor_idx, model.factor_names, as_idx=False)

    data = model_cache.uns[model_cache.UNS_GROUPED_R2].copy()
    groupby = data.columns[0]

    data = data[data["Factor"].isin(factor_idx)]

    data[model_cache.METRIC_R2] = data[
        [f"{model_cache.METRIC_R2}_{vn}" for vn in view_idx]
    ].sum(1)

    data = _subset_df(data, groupby, groups, include_rest=False)

    data = data.pivot(
        index="Factor",
        columns=groupby,
        values=model_cache.METRIC_R2,
    )
    data = data.loc[factor_idx]

    g = data.plot(
        kind=kind,
        stacked=stacked,
        color=kwargs.pop("color", _get_color_dict(model_cache.factor_adata, groupby)),
        **kwargs,
    )
    g.set_ylabel(r"$R^2$")

    g = _setup_legend(g)

    g.set_title(f"Variance explained across {groupby} groups in {', '.join(view_idx)}")
    g.set(xlabel="Factor")

    savefig_or_show("variance_explained_grouped", show=show, save=save)
    if not show:
        return g


def factors_overview(
    model,
    view_idx=0,
    one_sided=True,
    alpha=0.1,
    sig_only=False,
    prior_only=False,
    adjusted=False,
    top=25,
    show: Optional[bool] = None,
    save: Union[bool, str, None] = None,
    **kwargs,
):
    """Scatterplot of variance explained along with significance test results,
    see `muvi.tl.test`."""
    if isinstance(view_idx, int):
        view_idx = model.view_names[view_idx]

    model_cache = _get_model_cache(model)
    data = model_cache.factor_metadata.copy()

    name_col = "Factor"
    data[name_col] = data.index.astype(str)

    if prior_only:
        data = data.loc[~data[name_col].str.contains("dense", case=False), :].copy()

    size_col = None
    if model._informed:
        size_col = "Size"
        data[size_col] = model.get_prior_masks(view_idx, as_df=True)[view_idx].sum(1)
        data.loc[data[name_col].str.contains("dense", case=False), size_col] = 0

    sign_dict = {model_cache.TEST_ALL: " (*)"}
    if one_sided:
        sign_dict = {model_cache.TEST_NEG: " (-)", model_cache.TEST_POS: " (+)"}
    joint_p_col = "p_min"
    direction_col = "direction"

    p_col = "p"
    if adjusted:
        p_col = "p_adj"

    p_df = data[[f"{p_col}_{sign}_{view_idx}" for sign in sign_dict]]
    if p_df.isna().all(None) and sig_only:
        raise ValueError("No test results found in model cache, rerun `muvi.tl.test`.")

    p_df = p_df.fillna(1.0)
    data[joint_p_col] = p_df.min(axis=1).clip(1e-10, 1.0)
    data[direction_col] = p_df.idxmin(axis=1).str[len(p_col) + 1 : len(p_col) + 4]

    if alpha is None:
        alpha = 1.0
    if alpha <= 0:
        logger.warning("Negative or zero `alpha`, setting `alpha` to 0.01.")
        alpha = 0.01
    if alpha > 1.0:
        logger.warning("`alpha` larger than 1.0, setting `alpha` to 1.0.")
        alpha = 1.0
    data.loc[data[joint_p_col] > alpha, direction_col] = ""
    if sig_only:
        data = data.loc[data[direction_col] != "", :]

    data[name_col] = data[name_col] + data[direction_col].map(sign_dict).fillna("")

    neg_log_col = r"$-\log_{10}(FDR)$"
    data[neg_log_col] = -np.log10(data[joint_p_col])
    r2_col = f"r2_{view_idx}"
    if top > 0:
        data = data.sort_values(r2_col, ascending=True)

    g = sns.scatterplot(
        data=data.iloc[-top:],
        x=r2_col,
        y=name_col,
        hue=neg_log_col,
        palette=kwargs.pop("palette", "flare"),
        size=size_col,
        sizes=kwargs.pop("sizes", (50, 350)),
        **kwargs,
    )
    g.set_title(
        rf"Overview top factors in {view_idx} $\alpha = {alpha}$"
        # rf"($\alpha = {alpha:.{max(1, int(-np.log10(alpha)))}f}$)"  # noqa: ERA001
    )
    g.set(xlabel=r"$R^2$")
    savefig_or_show(f"overview_view_{view_idx}", show=show, save=save)
    if not show:
        return g


def inspect_factor(
    model,
    factor_idx,
    view_idx="all",
    top=25,
    ranked=True,
    figsize=None,
    show: Optional[bool] = None,
    save: Union[bool, str, None] = None,
    **kwargs,
):
    """Scatterplot of factor loadings for specific factors."""
    view_idx = _normalize_index(view_idx, model.view_names, as_idx=False)
    factor_idx = _normalize_index(factor_idx, model.factor_names, as_idx=False)

    n_views = len(view_idx)
    n_factors = len(factor_idx)

    if figsize is None:
        figsize = (8 * n_views, 8 * n_factors)

    fig, axs = plt.subplots(n_factors, n_views, squeeze=False, figsize=figsize)

    for m in range(n_views):
        view_name = view_idx[m]
        for k in range(n_factors):
            factor_name = factor_idx[k]
            i = k * n_views + m
            g = axs[k, m]
            # only last
            show_legend = i == n_views * n_factors - 1

            factor_loadings = model.get_factor_loadings(
                view_name, factor_name, as_df=True
            )[view_name].iloc[0, :]
            factor_mask = pd.Series(False, index=factor_loadings.index)
            if model._informed:
                factor_mask = model.get_prior_masks(view_name, factor_name, as_df=True)[
                    view_name
                ].iloc[0, :]
            else:
                logger.warning(
                    "Model not informed a priori, all features are inferred."
                )
            factor_loadings_abs = np.abs(factor_loadings)
            factor_loadings_rank = factor_loadings.rank(ascending=False)

            name_col = "Feature"
            loading_col = "Loading"
            rank_col = "Rank"
            abs_loading_col = "Loading (abs)"

            data = pd.DataFrame(
                {
                    name_col: model.feature_names[view_name],
                    "Mask": factor_mask,
                    loading_col: factor_loadings,
                    abs_loading_col: factor_loadings_abs,
                    rank_col: factor_loadings_rank,
                    "FP": ~factor_mask & factor_loadings_abs > 0.0,
                }
            )

            type_col = "Type"
            data[type_col] = data["FP"].map({False: "Annotated", True: "Inferred"})

            if top > 0:
                data = data.sort_values(abs_loading_col, ascending=True)
            x = loading_col
            y = name_col
            kwargs["hue"] = type_col
            kwargs["hue_order"] = ["Annotated", "Inferred"]
            kwargs["palette"] = {"Annotated": "black", "Inferred": "red"}
            if ranked:
                x = rank_col
                y = loading_col
            g = sns.scatterplot(
                ax=g,
                data=data.iloc[-top:],
                x=x,
                y=y,
                s=kwargs.pop("s", (64)),
                legend=show_legend and not ranked,
                **kwargs,
            )
            if ranked:
                g = sns.scatterplot(
                    ax=g,
                    data=data.iloc[:-top],
                    x=rank_col,
                    y=loading_col,
                    s=10,
                    legend=show_legend,
                    **kwargs,
                )

                y_max = factor_loadings.max()
                y_min = factor_loadings.min()
                x_range = factor_loadings_rank.max()

                labeled_data = (
                    data.iloc[-top:].sort_values(loading_col, ascending=False).copy()
                )

                labeled_data["is_positive"] = labeled_data[loading_col] > 0

                n_positive = labeled_data["is_positive"].sum()
                n_negative = top - n_positive
                num = max(n_positive, n_negative)

                labeled_data["x_arrow_pos"] = labeled_data[rank_col] + 0.02 * x_range
                labeled_data["x_text_pos"] = (
                    labeled_data["x_arrow_pos"] + 0.15 * x_range
                )
                labeled_data["y_arrow_pos"] = labeled_data[loading_col]
                labeled_data["y_text_pos"] = (
                    np.linspace(y_max, 0.1 * y_max, num=num)[:n_positive].tolist()
                    + np.linspace(y_min, -0.1 * y_min, num=num)[:n_negative][
                        ::-1
                    ].tolist()
                )

                for _, row in labeled_data.iterrows():
                    g.text(
                        row["x_text_pos"],
                        row["y_text_pos"],
                        row[name_col],
                        color=kwargs["palette"][row[type_col]],
                        fontsize="medium",
                    )
                    g.annotate(
                        "",
                        (row["x_arrow_pos"], row["y_arrow_pos"]),
                        xytext=(row["x_text_pos"], row["y_text_pos"]),
                        # bbox=dict(boxstyle="round", alpha=0.1),  # noqa: ERA001
                        arrowprops={
                            "arrowstyle": "simple,tail_width=0.01,head_width=0.15",
                            "color": "black",
                        },
                    )

            g.set_title(f"{factor_name} ({view_name})")

    fig.tight_layout()
    savefig_or_show("overview_factors", show=show, save=save)
    if not show:
        return fig, axs


# source: https://github.com/DTrimarchi10/confusion_matrix
def confusion_matrix(
    model,
    view_idx=0,
    true_mask=None,
    threshold=0.1,
    group_names=None,
    categories="auto",
    count=True,
    percent=True,
    cbar=True,
    xyticks=True,
    xyplotlabels=True,
    sum_stats=True,
    figsize=None,
    cmap="binary",
    title=None,
):
    if true_mask is None:
        view_idx, true_mask = model.get_prior_masks(view_idx).popitem()
    cf = sklearn_cf(
        true_mask.flatten(),
        (
            np.abs(model.get_factor_loadings(view_idx).popitem()[1]).flatten()
            > threshold
        ),
    )
    if group_names is None:
        group_names = ["true neg", "false pos", "false neg", "true pos"]
    if categories is None:
        categories = ["0.0", "1.0"]
    # CODE TO GENERATE TEXT INSIDE EACH SQUARE
    blanks = ["" for i in range(cf.size)]

    if group_names and len(group_names) == cf.size:
        group_labels = [f"{value}\n" for value in group_names]
    else:
        group_labels = blanks

    group_counts = [f"{value:0.0f}\n" for value in cf.flatten()] if count else blanks

    if percent:
        group_percentages = [f"{value:.2%}" for value in cf.flatten() / np.sum(cf)]
    else:
        group_percentages = blanks

    box_labels = [
        f"{v1}{v2}{v3}".strip()
        for v1, v2, v3 in zip(group_labels, group_counts, group_percentages)
    ]
    box_labels = np.asarray(box_labels).reshape(cf.shape[0], cf.shape[1])

    # CODE TO GENERATE SUMMARY STATISTICS & TEXT FOR SUMMARY STATS
    if sum_stats:
        # Accuracy is sum of diagonal divided by total observations
        accuracy = np.trace(cf) / float(np.sum(cf))

        # if it is a binary confusion matrix, show some more stats
        if len(cf) == 2:
            # Metrics for Binary Confusion Matrices
            precision = cf[1, 1] / sum(cf[:, 1])
            recall = cf[1, 1] / sum(cf[1, :])
            f1_score = 2 * precision * recall / (precision + recall)
            stats_text = (
                f"\n\nAccuracy={accuracy:0.3f}"
                + "\nPrecision={:0.3f}\nRecall={:0.3f}\nF1 Score={:0.3f}".format(
                    precision, recall, f1_score
                )
            )
        else:
            stats_text = f"\n\nAccuracy={accuracy:0.3f}"
    else:
        stats_text = ""

    # SET FIGURE PARAMETERS ACCORDING TO OTHER ARGUMENTS
    if figsize is None:
        # Get default figure size if not set
        figsize = plt.rcParams.get("figure.figsize")

    if not xyticks:
        # Do not show categories if xyticks is False
        categories = False

    # MAKE THE HEATMAP VISUALIZATION
    plt.figure(figsize=figsize)
    g = sns.heatmap(
        cf,
        annot=box_labels,
        fmt="",
        cmap=cmap,
        cbar=cbar,
        xticklabels=categories,
        yticklabels=categories,
    )

    if xyplotlabels:
        plt.ylabel("True label")
        plt.xlabel("Predicted label" + stats_text)
    else:
        plt.xlabel(stats_text)

    if title is None:
        title = f"View {view_idx}"
    plt.title(title)

    return g


def factor_activity(
    true_w,
    approx_w,
    true_mask,
    noisy_mask,
    factor_idx=0,
    ylim=None,
    **kwargs,
):
    true_w_col = true_w[factor_idx, :]
    w_col = approx_w[factor_idx, :]
    true_mask_col = true_mask[factor_idx, :]
    noisy_mask_col = noisy_mask[factor_idx, :]

    activity_df = pd.DataFrame(
        {
            "true_weight": true_w_col,
            "weight": w_col,
            "true_mask": true_mask_col,
            "noisy_mask": noisy_mask_col,
            "TP": true_mask_col * noisy_mask_col,
            "FP": (1 - true_mask_col) * noisy_mask_col,
            "TN": (1 - true_mask_col) * (1 - noisy_mask_col),
            "FN": true_mask_col * (1 - noisy_mask_col),
        }
    )
    activity_df.sort_values(["true_weight"], inplace=True)

    score_cols = ["TP", "FP", "TN", "FN"]

    assert (activity_df.loc[:, score_cols].values.sum(1) == 1).all()
    activity_df["state"] = (
        activity_df.loc[:, score_cols]
        .astype(np.int32)
        .dot(activity_df.loc[:, score_cols].columns + "+")
        .str[:-1]
    )
    activity_df["true state"] = [
        "on" if f > 0.5 else "off" for f in activity_df["true_mask"]
    ]
    activity_df["idx"] = list(range(len(w_col)))

    g = sns.scatterplot(
        data=activity_df,
        x="idx",
        y="weight",
        hue="state",
        hue_order=["TP", "FN", "TN", "FP"],
        style="true state",
        style_order=["on", "off"],
        size="state",
        sizes={"TP": 64, "FN": 64, "TN": 32, "FP": 32},
        linewidth=0.01,
        **kwargs,
    )
    g.set_xlabel("")
    joint_handles, joint_labels = g.get_legend_handles_labels()
    g.legend(
        handles=[h for i, h in enumerate(joint_handles) if i not in [0, 5]],
        labels=[h for i, h in enumerate(joint_labels) if i not in [0, 5]],
    )
    if ylim is not None:
        g.set(ylim=ylim)

    return g, activity_df


# scanpy plotting


# plot latent embeddings
def _embed(model, color, pl_fn, **kwargs):
    return pl_fn(_get_model_cache(model).factor_adata, color=color, **kwargs)


def tsne(model, color, **kwargs):
    if "X_tsne" not in _get_model_cache(model).factor_adata.obsm:
        logger.warning("Computing tSNE embeddings first.")
        muvi.tl.tsne(model)
    return _embed(model, color, pl_fn=sc.pl.tsne, **kwargs)


def umap(model, color, **kwargs):
    if "X_umap" not in _get_model_cache(model).factor_adata.obsm:
        logger.warning("Computing UMAP embeddings first.")
        muvi.tl.umap(model)
    return _embed(model, color, pl_fn=sc.pl.umap, **kwargs)


# plot groups of observations against (subset of) factors
def group(
    model,
    factor_idx,
    groupby,
    groups=None,
    pl_type=HEATMAP,
    **kwargs,
):
    pl_type = pl_type.lower().strip()

    if (pl_type in MATRIXPLOT or pl_type in DOTPLOT) and "colorbar_title" not in kwargs:
        kwargs["colorbar_title"] = "Average scores\nin group"

    if pl_type in DOTPLOT and "size_title" not in kwargs:
        kwargs["size_title"] = "Fraction of samples\nin group (%)"

    type_to_fn = {
        HEATMAP: sc.pl.heatmap,
        MATRIXPLOT: sc.pl.matrixplot,
        DOTPLOT: sc.pl.dotplot,
        TRACKSPLOT: sc.pl.tracksplot,
        VIOLIN: sc.pl.violin,
        STACKED_VIOLIN: sc.pl.stacked_violin,
    }

    factor_idx = _normalize_index(factor_idx, model.factor_names, as_idx=False)

    try:
        pl_fn = type_to_fn[pl_type]
    except KeyError as e:
        raise ValueError(
            f"`{pl_type}` is not valid. Select one of {','.join(PL_TYPES)}."
        ) from e

    adata = _get_model_cache(model).factor_adata

    return pl_fn(
        adata[
            _subset_df(adata.obs.copy(), groupby, groups, include_rest=False).index, :
        ],
        factor_idx,
        groupby,
        **kwargs,
    )


# plot ranked factors against groups of observations
def rank(model, n_factors=10, pl_type=None, sep_groups=True, **kwargs):
    factor_adata = _get_model_cache(model).factor_adata
    if "rank_genes_groups" not in factor_adata.uns:
        raise ValueError("No group-wise ranking found, run `muvi.tl.rank first.`")

    groupby = factor_adata.uns["rank_genes_groups"]["params"]["groupby"]
    dendrogram_key = "dendrogram_" + groupby
    if dendrogram_key not in factor_adata.uns:
        logger.warning(
            f"dendrogram data not found (using `{dendrogram_key}` as key). "
            "Running `muvi.tl.dendrogram` with default parameters. "
            "For fine tuning "
            "it is recommended to run `muvi.tl.dendrogram` independently."
        )
        muvi.tl.dendrogram(model, groupby)

    if pl_type is None:
        logger.warning("`pl_type` is None, defaulting to `sc.pl.rank_genes_groups`.")
        pl_type = ""
    pl_type = pl_type.lower().strip()
    n_factors = kwargs.pop("n_genes", n_factors)

    type_to_fn = {
        "": sc.pl.rank_genes_groups,
        HEATMAP: sc.pl.rank_genes_groups_heatmap,
        MATRIXPLOT: sc.pl.rank_genes_groups_matrixplot,
        DOTPLOT: sc.pl.rank_genes_groups_dotplot,
        TRACKSPLOT: sc.pl.rank_genes_groups_tracksplot,
        VIOLIN: sc.pl.rank_genes_groups_violin,
        STACKED_VIOLIN: sc.pl.rank_genes_groups_stacked_violin,
    }

    n_groups = len(
        factor_adata.obs[
            factor_adata.uns["rank_genes_groups"]["params"]["groupby"]
        ].unique()
    )
    if "groups" in kwargs and kwargs["groups"] is not None:
        n_groups = len(kwargs["groups"])

    positions = np.linspace(
        n_factors, n_factors * n_groups, num=n_groups - 1, endpoint=False
    )

    try:
        pl_fn = type_to_fn[pl_type]
    except KeyError as e:
        raise ValueError(
            f"`{pl_type}` is not valid. Select one of {', '.join(PL_TYPES)}."
        ) from e

    if not sep_groups:
        return pl_fn(
            factor_adata,
            n_genes=n_factors,
            **kwargs,
        )

    show = kwargs.pop("show", None)
    save = kwargs.pop("save", None)
    _pl = pl_fn(
        factor_adata,
        n_genes=n_factors,
        show=False,
        save=None,
        **kwargs,
    )

    # add line separation
    g = None
    if pl_type == HEATMAP:
        g = _pl["heatmap_ax"]
        ymin = -0.5
        ymax = factor_adata.n_obs
        positions -= 0.5
    if pl_type == MATRIXPLOT:
        g = _pl["mainplot_ax"]
        ymin = 0.0
        ymax = n_groups
    if pl_type in (DOTPLOT, STACKED_VIOLIN):
        g = _pl["mainplot_ax"]
        ymin = -0.5
        ymax = n_groups + 0.5

    if g is not None:
        g = _lines(
            g,
            positions,
            ymin=ymin,
            ymax=ymax,
            horizontal=kwargs.pop("swap_axes", False),
            lw=0.5,
            color="black",
            linestyles="dashed",
            zorder=10,
            clip_on=False,
        )
    writekey = "rank"
    if len(pl_type) > 0:
        writekey += f"_{pl_type}"
    savefig_or_show(writekey, show=show, save=save)
    if not show:
        return g


def clustermap(model, factor_idx="all", **kwargs):
    factor_idx = _normalize_index(factor_idx, model.factor_names, as_idx=False)
    return sc.pl.clustermap(
        _get_model_cache(model).factor_adata[:, factor_idx], **kwargs
    )


def stripplot(
    model,
    factor_idx,
    groupby,
    groups=None,
    include_rest=True,
    rot: int = 45,
    show: Optional[bool] = None,
    save: Union[bool, str, None] = None,
    **kwargs,
):
    factor_idx = _normalize_index(factor_idx, model.factor_names, as_idx=False)

    model_cache = _get_model_cache(model)
    if groupby not in model_cache.factor_adata.obs.columns:
        raise ValueError(
            f"`{groupby}` not found in the metadata, "
            " add a new column onto `model._cache.factor_adata.obs`."
        )
    data = pd.concat(
        [
            model_cache.factor_adata.to_df().loc[:, factor_idx],
            model_cache.factor_adata.obs[groupby],
        ],
        axis=1,
    )

    data = pd.melt(data, id_vars=[groupby], var_name="Factor", value_name="Score")

    data = _subset_df(data, groupby, groups, include_rest=include_rest)

    g = sns.stripplot(
        data=data,
        x="Factor",
        y="Score",
        hue=kwargs.pop("hue", groupby),
        palette=kwargs.pop(
            "palette",
            _get_color_dict(
                model_cache.factor_adata, groupby, include_rest=include_rest
            ),
        ),
        dodge=kwargs.pop("dodge", True),
        **kwargs,
    )
    if rot is not None:
        for label in g.get_xticklabels():
            label.set_rotation(rot)
    g = _setup_legend(g, remove_last=groups is not None and include_rest)
    savefig_or_show("stripplot", show=show, save=save)
    if not show:
        return g


def scatter(
    model,
    x,
    y,
    groupby=None,
    groups=None,
    include_rest=True,
    style=None,
    markers=True,
    **kwargs,
):
    x = _normalize_index(x, model.factor_names, as_idx=False)[0]
    y = _normalize_index(y, model.factor_names, as_idx=False)[0]
    kwargs["color"] = groupby
    model_cache = _get_model_cache(model)

    data = pd.concat(
        [model_cache.factor_adata.to_df(), model_cache.factor_adata.obs.copy()], axis=1
    )

    data = _subset_df(data, groupby, groups, include_rest=include_rest)
    palette = kwargs.pop(
        "palette",
        _get_color_dict(model_cache.factor_adata, groupby, include_rest=include_rest),
    )

    if style is None:
        adata = model_cache.factor_adata
        if include_rest and groups is not None:
            groups = groups[:-1]
        else:
            adata = adata[data.index, :]
        return sc.pl.scatter(
            adata,
            x,
            y,
            groups=groups,
            **kwargs,
        )

    logger.warning(
        "Experimental! "
        "Passing a `style` argument does not rely on `sc.pl.scatter`, "
        "and may lead to undesired results."
    )

    size = kwargs.pop("size", None)
    show = kwargs.pop("show", None)
    save = kwargs.pop("save", None)

    kwargs = {}

    if size is None:
        size = 120000 / data.shape[0]
    g = sns.scatterplot(
        data=data,
        x=x,
        y=y,
        hue=groupby,
        style=style,
        markers=markers,
        s=size,
        palette=palette,
        linewidth=kwargs.pop("linewidth", 0),
        ax=kwargs.pop("ax", None),
        **kwargs,
    )

    # getting as close as possible to scanpy plotting style
    g = _setup_legend(g, remove_last=groups is not None and include_rest)

    g.set_title(groupby)
    savefig_or_show("scatter", show=show, save=save)
    if not show:
        return g


def scatter_rank(model, groupby=None, groups=None, **kwargs):
    group_df = sc.get.rank_genes_groups_df(model._cache.factor_adata, group=groups)
    group_df["scores_abs"] = group_df["scores"].abs()

    relevant_factors_dict = {}
    for group in group_df["group"].unique():
        relevant_factors_dict[group] = (
            group_df[group_df["group"] == group]
            .sort_values("scores_abs", ascending=False)
            .iloc[:2]["names"]
            .tolist()
        )

    show = kwargs.pop("show", None)
    save = kwargs.pop("save", None)
    gs = {}

    for group, relevant_factors in relevant_factors_dict.items():
        g = muvi.pl.scatter(
            model,
            *relevant_factors[:2],
            groupby=groupby,
            groups=groups,
            show=False,
            save=False,
            **kwargs,
        )

        g.set_title(f"{groupby} ({group})")
        savefig_or_show(f"scatter_rank_{group}", show=show, save=save)
        gs[group] = g
    if not show:
        return gs



# Highly inspired by https://github.com/krassowski/gsea-api
import logging

from collections import Counter
from pathlib import Path
from typing import Collection
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import Set
from typing import Tuple

import numpy as np
import pandas as pd

from sklearn.metrics import pairwise_distances


logger = logging.getLogger(__name__)


class FeatureSet:
    def __init__(
        self,
        features: Collection[str],
        name: str,
        description: str = "",
    ):
        self.name = name
        self.features = frozenset(features)
        self.description = description

        if self.empty:
            logger.warning(f"FeatureSet {name!r} is empty.")

        redundant_features = None

        if len(features) != len(self.features):
            redundant_features = {
                feature: count
                for feature, count in Counter(features).items()
                if count > 1
            }

            logger.warning(
                f"FeatureSet {name!r} received a non-unique "
                f"collection of features; redundant features: {redundant_features}"
            )

        self.redundant_features = redundant_features

    @property
    def empty(self):
        return len(self) == 0

    def __len__(self):
        return len(self.features)

    def __repr__(self) -> str:
        features = (
            ": " + ", ".join(sorted(self.features)) if len(self.features) < 5 else ""
        )
        return f"<FeatureSet {self.name!r} with {len(self)} features{features}>"

    def __iter__(self) -> Iterable[str]:
        return iter(self.features)

    def __eq__(self, other: "FeatureSet") -> bool:
        return self.features == other.features

    def __hash__(self) -> int:
        return hash(self.features)

    def __and__(self, other: "FeatureSet") -> "FeatureSet":
        return FeatureSet(
            self.features & other.features,
            name=f"{self.name}&{other.name}",
        )

    def __or__(self, other: "FeatureSet") -> "FeatureSet":
        return FeatureSet(
            self.features | other.features,
            name=f"{self.name}|{other.name}",
        )

    def __add__(self, other: "FeatureSet") -> "FeatureSet":
        return self.__or__(other)

    def subset(self, features: Iterable[str]) -> "FeatureSet":
        """Subset features from a feature set.

        Parameters
        ----------
        features : Iterable[str]
            Features to subset.

        Returns
        -------
        FeatureSet
            A new feature set with the subset of features.
        """
        return FeatureSet(
            self.features & features,
            name=self.name,
        )


class FeatureSets:
    def __init__(
        self,
        feature_sets: Collection[FeatureSet],
        name: str = "UNL",
        remove_empty: bool = True,
    ):
        self.name = name

        if remove_empty:
            feature_sets = {
                feature_set for feature_set in feature_sets if not feature_set.empty
            }

        redundant_feature_sets = None

        if len(set(feature_sets)) != len(feature_sets):
            redundant_feature_sets = {
                feature_set: count
                for feature_set, count in Counter(feature_sets).items()
                if count > 1
            }

            logger.warning(
                f"FeatureSets {name!r} received a non-unique "
                "collection of feature sets; redundant feature sets: "
                f"{redundant_feature_sets}"
            )
        self.feature_sets = frozenset(feature_sets)
        self.redundant_feature_sets = redundant_feature_sets

    @property
    def empty(self):
        return len(self) == 0

    @property
    def median(self) -> int:
        return int(np.median([len(fs) for fs in self.feature_sets]))

    @property
    def features(self) -> frozenset:
        return frozenset.union(*[fs.features for fs in self.feature_sets])

    @property
    def feature_set_by_name(self) -> dict:
        return {feature_set.name: feature_set for feature_set in self.feature_sets}

    def __getitem__(self, name: str) -> FeatureSet:
        return self.feature_set_by_name[name]

    def __len__(self):
        return len(self.feature_sets)

    def __iter__(self) -> Iterable[FeatureSet]:
        return iter(self.feature_sets)

    def __repr__(self) -> str:
        feature_sets = (
            ": " + ", ".join(sorted({fs.name for fs in self.feature_sets}))
            if len(self.feature_sets) < 5
            else ""
        )
        return (
            f"<FeatureSets {self.name!r} with {len(self)} "
            + f"feature sets{feature_sets}>"
        )

    def __eq__(self, other: "FeatureSets") -> bool:
        return self.feature_sets == other.feature_sets

    def __hash__(self) -> int:
        return hash(self.feature_sets)

    def __and__(self, other: "FeatureSets") -> "FeatureSets":
        return FeatureSets(
            name=f"{self.name}&{other.name}",
            feature_sets=self.feature_sets & other.feature_sets,
        )

    def __or__(self, other: "FeatureSets") -> "FeatureSets":
        return FeatureSets(
            name=f"{self.name}|{other.name}",
            feature_sets=self.feature_sets | other.feature_sets,
        )

    def __add__(self, other: "FeatureSets") -> "FeatureSets":
        return self.__or__(other)

    def find(self, partial_name: str):
        """Perform a simple search given a (partial) feature set name.

        Parameters
        ----------
        partial_name : str
            Feature set (partial) name to search for.

        Returns
        -------
        FeatureSets
            Search results.
        """
        return FeatureSets(
            {
                feature_set
                for feature_set in self.feature_sets
                if partial_name in feature_set.name
            },
            name=f"{self.name}:{partial_name}",
        )

    def remove(self, names: Iterable[str]):
        """Remove feature sets by name.

        Parameters
        ----------
        names : Iterable[str]
            Collection of feature set names.
        """
        return FeatureSets(
            {
                feature_set
                for feature_set in self.feature_sets
                if feature_set.name not in names
            },
            name=self.name,
        )

    def keep(self, names: Iterable[str]):
        """Keep feature sets by name.

        Parameters
        ----------
        names : Iterable[str]
            Collection of feature set names.
        """
        return FeatureSets(
            {
                feature_set
                for feature_set in self.feature_sets
                if feature_set.name in names
            },
            name=self.name,
        )

    def trim(self, min_count: int = 1, max_count: Optional[int] = None):
        """Trim feature sets by min/max size.

        Parameters
        ----------
        min_count : int, optional
            Minimum number of features, by default 1.
        max_count : int, optional
            Maximum number of features, by default None.
        """
        return FeatureSets(
            {
                feature_set
                for feature_set in self.feature_sets
                if min_count
                <= len(feature_set.features)
                <= (max_count or len(feature_set.features))
            },
            name=self.name,
        )

    def subset(self, features: Iterable[str]):
        """Subset feature sets by features.

        Parameters
        ----------
        features : Iterable[str]
            Collection of features.
        """
        return FeatureSets(
            {feature_set.subset(set(features)) for feature_set in self.feature_sets},
            name=self.name,
        )

    def filter(
        self,
        features: Iterable[str],
        min_fraction: float = 0.5,
        min_count: int = 1,
        keep: Optional[Iterable[str]] = None,
        subset: bool = True,
    ):
        """Filter feature sets.

        Parameters
        ----------
        features : Iterable[str]
            Features to filter.
        min_fraction : float, optional
            Mininimum portion of the feature set to be present in `features`,
            by default 0.5
        min_count : int, optional
            Minimum size of the intersection set
            between a feature set and the set of `features`,
            by default 1
        keep : Iterable[str], optional
            Feature sets to keep regardless of the filter conditions,
            by default None
        subset : bool, optional
            Whether to subset the resulting feature sets based on `features`,
            by default True

        Returns
        -------
        FeatureSets
            Filtered feature sets.
        """
        features = set(features)

        if keep is None:
            keep = set()

        feature_set_subset = set()

        for feature_set in self.feature_sets:
            if feature_set.name in keep:
                feature_set_subset.add(feature_set)
                continue
            intersection = features & feature_set.features
            count = len(intersection)
            fraction = count / len(feature_set)
            if count >= min_count and fraction >= min_fraction:
                feature_set_subset.add(feature_set)

        filtered_feature_sets = FeatureSets(feature_set_subset, name=self.name)
        if subset:
            filtered_feature_sets = filtered_feature_sets.subset(features)
        return filtered_feature_sets

    def to_mask(
        self, features: Optional[Iterable[str]] = None, sort: bool = True
    ) -> pd.DataFrame:
        """Convert feature sets to a mask.

        Parameters
        ----------
        features : Iterable[str], optional
            Collection of features, by default None.
        sort : bool, optional
            Sort feature sets alphabetically, by default True.

        Returns
        -------
        pd.DataFrame
            Mask of features.
        """
        features = features or self.features
        features_list = list(features)
        feature_sets_list = list(self.feature_sets)
        if sort:
            feature_sets_list = sorted(feature_sets_list, key=lambda fs: fs.name)
        return pd.DataFrame(
            [
                [feature in feature_set.features for feature in features_list]
                for feature_set in feature_sets_list
            ],
            index=[feature_set.name for feature_set in feature_sets_list],
            columns=features_list,
        )

    def similarity_to_feature_sets(
        self, other: "FeatureSets" = None, metric: str = "jaccard"
    ) -> pd.DataFrame:
        """Compute similarity matrix between feature sets.

        Parameters
        ----------
        other : FeatureSets, optional
            Other feature set collection, by default None.
        metric : str, optional
            Similarity metric, by default "jaccard".

        Returns
        -------
        pd.DataFrame
            Similarity matrix as 1 minus distance matrix,
            may lead to negative values for some distance metrics.
        """

        if metric not in ["jaccard", "cosine"]:
            logger.warning(
                f"Similarity matrix for `{metric}` might be negative. "
                "Recommended metrics are `jaccard` or `cosine`."
            )

        self_mask = self.to_mask()
        other_mask = other.to_mask() if other else self_mask
        return 1 - pd.DataFrame(
            pairwise_distances(
                self_mask.to_numpy(), other_mask.to_numpy(), metric=metric
            ),
            index=self_mask.index,
            columns=other_mask.index,
        )

    def similarity_to_observations(
        self,
        observations: pd.DataFrame,
    ) -> pd.DataFrame:
        """Compute similarity matrix between feature sets.

        Parameters
        ----------
        observations : pd.DataFrame
            Dataframe of observations.

        Returns
        -------
        pd.DataFrame
            Similarity matrix as correlation matrix.
        """
        obs_mean = observations.mean(axis=1)

        dist_to_mean_dict = {}
        for feature_set in self.feature_sets:
            col_subset = [
                col for col in observations.columns if col in feature_set.features
            ]
            if len(col_subset) == 0:
                dist_to_mean_dict[feature_set.name] = pd.Series(
                    np.nan, index=observations.index
                )
                continue
            dist_to_mean_dict[feature_set.name] = (
                observations.loc[:, col_subset].mean(axis=1) - obs_mean
            )

        return pd.DataFrame(dist_to_mean_dict).corr()

    def _find_similar_pairs(
        self, sim_matrix: pd.DataFrame, similarity_threshold: float
    ) -> Set[Tuple[str, str]]:
        """Find similar pairs of feature sets.

        Parameters
        ----------
        sim_matrix : pd.DataFrame
            Similarity matrix.
        similarity_threshold : float
            Similarity threshold to consider similar pairs.

        Returns
        -------
        Set[Tuple[str, str]]
            Similar pairs of feature sets.
        """

        pairs = set()
        visited = set()

        row_offset = 0
        for current_fs, row in sim_matrix.iterrows():
            row_offset += 1
            if row_offset >= len(row):
                break
            if current_fs in visited:
                continue
            visited.add(current_fs)
            closest_fs = row.iloc[row_offset:].idxmax()
            similarity = row[closest_fs]
            if similarity >= similarity_threshold and closest_fs not in visited:
                pairs.add((current_fs, closest_fs, similarity))
                visited.add(closest_fs)

        return pairs

    def find_similar_pairs(
        self,
        observations: pd.DataFrame = None,
        metric: Optional[str] = None,
        similarity_threshold: float = 0.8,
    ) -> Set[Tuple[str, str]]:
        """Find similar pairs of feature sets.

        Parameters
        ----------
        observations : pd.DataFrame, optional
            Dataframe of observations, if provided, the similarity between feature sets
            is computed based on the correlation of the similarity from the mean
            of the observations in the feature set, by default None.
        metric : str, optional
            Similarity metric, by default "jaccard" if observations not provided.
        similarity_threshold : float, optional
            Similarity threshold to consider similar pairs,
            by default 0.8.

        Returns
        -------
        Set[Tuple[str, str]]
            Similar pairs of feature sets.
        """
        if observations is None and metric is None:
            logger.warning(
                "Neither observations nor metric is provided,"
                " using `metric=jaccard` as default."
            )
            metric = "jaccard"

        sim_matrix = []
        if observations is not None:
            sim_matrix.append(self.similarity_to_observations(observations))
        if metric is not None:
            sim_matrix.append(self.similarity_to_feature_sets(metric=metric))

        if observations is not None and metric is not None:
            sim_matrix[0][sim_matrix[0] < 0] = 0.0
            sim_matrix[1][sim_matrix[1] < 0] = 0.0
            sim_matrix = (2 * sim_matrix[0] * sim_matrix[1]) / (
                sim_matrix[0] + sim_matrix[1]
            )
        else:
            sim_matrix = sim_matrix[0]
        return self._find_similar_pairs(sim_matrix.fillna(0.0), similarity_threshold)

    def merge_pairs(self, pairs: Iterable[Tuple[str, str]]):
        """Merge pairs of feature sets.

        Parameters
        ----------
        pairs : Iterable[Tuple[str, str]]
            Pairs of feature sets.

        Returns
        -------
        FeatureSets
            Merged feature sets.
        """
        names_to_remove = set()
        merged_feature_sets = set()
        for pair in pairs:
            merged_feature_sets.add(self[pair[0]] | self[pair[1]])
            names_to_remove |= {pair[0], pair[1]}

        # remove merged feature sets
        feature_sets = self.remove(names_to_remove)
        # then add merged feature sets
        feature_sets |= FeatureSets(merged_feature_sets)
        feature_sets.name = self.name
        return feature_sets

    def merge_similar(
        self,
        observations: pd.DataFrame = None,
        metric: Optional[str] = None,
        similarity_threshold: float = 0.8,
        iteratively: bool = True,
    ):
        """Merge similar feature sets.


        Parameters
        ----------
        observations : pd.DataFrame, optional
            Dataframe of observations, if provided, the similarity between feature sets
            is computed based on the correlation of the similarity from the mean
            of the observations in the feature set, by default None.
        metric : str, optional
            Similarity metric, by default "jaccard" if observations not provided.
        similarity_threshold : float, optional
            Similarity threshold to consider similar pairs,
            by default 0.8.
        iteratively : bool, optional
            Whether to merge iteratively, by default True

        Returns
        -------
        FeatureSets
            Merged feature sets.
        """

        feature_sets = self
        while True:
            pairs = {
                (name1, name2)
                for name1, name2, _ in feature_sets.find_similar_pairs(
                    observations=observations,
                    metric=metric,
                    similarity_threshold=similarity_threshold,
                )
            }
            stopping = ""
            if len(pairs) == 0 and iteratively:
                stopping = " Stopping..."
            logger.info(f"Found {len(pairs)} pairs to merge.{stopping}")
            feature_sets = feature_sets.merge_pairs(pairs)
            if len(pairs) == 0 or not iteratively:
                break
        return feature_sets

    def to_gmt(self, path: Path):
        """Write this feature set collection to a GMT file.

        Parameters
        ----------
        path : Path
            Path to the output file.
        """
        with open(path, "w") as f:
            for feature_set in self.feature_sets:
                f.write(
                    feature_set.name
                    + "\t"
                    + feature_set.description
                    + "\t"
                    + "\t".join(feature_set.features)
                    + "\n"
                )

    def to_dict(self) -> Dict[str, Iterable[str]]:
        """Convert this feature set collection to a dictionary.

        Returns
        -------
        Dict[str, Iterable[str]]
            Dictionary of feature sets.
        """
        return {fs.name: fs.features for fs in self.feature_sets}


def from_gmt(path: Path, name: Optional[str] = None, **kwargs) -> FeatureSets:
    """Create a FeatureSets object from a GMT file.

    Parameters
    ----------
    path : Path
        Path to the GMT file.
    name : str, optional
        Name of the collection, by default None.

    Returns
    -------
    FeatureSets
    """
    feature_sets = set()
    with open(path) as f:
        for line in f:
            fs_name, description, *features = line.strip().split("\t")
            feature_sets.add(
                FeatureSet(
                    features,
                    name=fs_name,
                    description=description,
                )
            )
    return FeatureSets(feature_sets, name=name or Path(path).name, **kwargs)


def from_dict(
    d: Dict[str, Iterable[str]],
    name: Optional[str] = None,
    **kwargs,
) -> FeatureSets:
    """Create a FeatureSets object from a dictionary.

    Parameters
    ----------
    d : Dict[str, Iterable[str]]
        Dictionary of feature sets.
    name : str, optional
        Name of the collection, by default None.

    Returns
    -------
    FeatureSets
    """
    feature_sets = set()
    for fs_name, features in d.items():
        feature_sets.add(FeatureSet(features, name=fs_name))
    return FeatureSets(feature_sets, name=name, **kwargs)


def from_dataframe(
    df: pd.DataFrame,
    name: Optional[str] = None,
    name_col: str = "name",
    features_col: str = "features",
    desc_col: Optional[str] = None,
    **kwargs,
) -> FeatureSets:
    """Create a FeatureSets object from a DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame of feature sets.
    name : str, optional
        Name of the collection, by default None.
    name_col : str, optional
        Name of the column containing feature set names, by default "name".
    features_col : str, optional
        Name of the column containing feature set features, by default "features".
    desc_col : str, optional
        Name of the column containing feature set descriptions, by default None.

    Returns
    -------
    FeatureSets
    """
    feature_sets = set()
    for _, row in df.iterrows():
        feature_sets.add(
            FeatureSet(
                row[features_col],
                name=row[name_col],
                description=desc_col is not None and row[desc_col],
            )
        )
    return FeatureSets(feature_sets, name=name, **kwargs)



import logging

import numpy as np


logger = logging.getLogger(__name__)


def _normalize_index(indexer, index, as_idx=True):
    # work with ints, convert at the end
    # if single str, get idx and put to list
    # TODO: can be an issue if any of the indices is named 'all'..
    if indexer is None:
        raise IndexError("None index.")
    if isinstance(indexer, str):
        indexer = range(len(index)) if indexer == "all" else [index.get_loc(indexer)]
    # if single integer, put to list
    if isinstance(indexer, (np.integer, int)):
        indexer = [indexer]
    # work with np array
    indexer = np.array(indexer)
    # if empty
    if len(indexer) == 0:
        raise IndexError(f"Empty index, `{indexer}`.")
    # if mask, get indices where True
    if isinstance(indexer[0], (bool, np.bool_)):
        indexer = np.where(indexer)[0]
    # if all False from previous boolean mask
    if len(indexer) == 0:
        raise IndexError(f"Empty index, `{indexer}`.")
    # if str, get indices where names match
    if isinstance(indexer[0], (str, np.str_)):
        _indexer = index.get_indexer(indexer)
        bad_idx = [idx for _idx, idx in zip(_indexer, indexer) if _idx == -1]
        if len(bad_idx) > 0:
            logger.warning(f"Invalid index, `{bad_idx}`, removing...")
        indexer = [_idx for _idx in _indexer if _idx != -1]
    # if all bad indices
    if len(indexer) == 0:
        raise IndexError(f"Empty index, `{indexer}`.")
    if isinstance(indexer[0], (int, np.integer)):
        if as_idx:
            return indexer
        return index[indexer]
    raise IndexError(f"Invalid index, `{indexer}`.")



"""Collection of MuVI callbacks."""
from collections import defaultdict
from typing import Callable
from typing import List
from typing import Optional

import numpy as np

from sklearn.metrics import precision_recall_fscore_support

from muvi.tools.utils import rmse
from muvi.tools.utils import variance_explained


class EarlyStoppingCallback:
    def __init__(
        self,
        n_epochs: int,
        min_epochs: int = 100,
        tolerance: float = 1e-5,
        patience: int = 10,
    ):
        """Early stopping callback.

        Parameters
        ----------
        n_epochs : int
            Number of training epochs
        min_epochs : int, optional
            Minimal number of epochs before deploying early stopping,
            by default 100
        tolerance : float, optional
            Improvement ratio between two consecutive evaluations,
            by default 1e-5
        patience : int, optional
            Number of patience steps before terminating training,
            by default 10
        """
        self.n_epochs = n_epochs
        self.min_epochs = min_epochs
        self.tolerance = tolerance
        self.patience = patience

        self.early_stop_counter = 0
        self.min_loss = np.inf

    def __call__(self, loss_history):
        epoch_idx = len(loss_history) - 1
        if epoch_idx < self.min_epochs:
            return False

        current_loss = loss_history[epoch_idx]
        if self.min_loss == np.inf:
            self.min_loss = current_loss
            return False

        stop_early = False

        relative_improvement = (self.min_loss - current_loss) / np.abs(self.min_loss)
        self.min_loss = min(self.min_loss, current_loss)

        if relative_improvement < self.tolerance:
            self.early_stop_counter += 1
        else:
            self.early_stop_counter = 0
        stop_early = self.early_stop_counter >= self.patience

        if stop_early:
            print(
                "Relative improvement of "
                f"{relative_improvement:0.4g} < {self.tolerance:0.4g} "
                f"for {self.patience} step(s) in a row, stopping early."
            )

        return stop_early


class CheckpointCallback:
    def __init__(
        self,
        model,
        n_epochs: int,
        n_checkpoints: int = 10,
        callback: Optional[Callable] = None,
    ):
        """Checkpoint callback.

        Parameters
        ----------
        model : MuVI
            A MuVI object
        n_epochs : int
            Number of training epochs
        n_checkpoints : int, optional
            Number of times to execute during training,
            by default 10
        callback : Callable, optional
            A function that accepts the loss history,
            by default None
        """
        self.model = model
        self.n_epochs = n_epochs
        self.n_checkpoints = n_checkpoints
        self.window_size = n_epochs // n_checkpoints
        self.callback = callback

    def __call__(self, loss_history):
        epoch_idx = len(loss_history)

        if epoch_idx % self.window_size == 0 or epoch_idx == self.n_epochs:
            self.callback()

        return False


class LogCallback:
    from typing import ClassVar

    _CALLBACK_KEYS: ClassVar[list] = [
        "sparsity",
        "binary_scores",
        "rmse",
        "r2",
        "log",
    ]

    def __init__(
        self,
        model,
        n_epochs: int,
        n_checkpoints: int = 10,
        active_callbacks: Optional[List[str]] = None,
        view_wise: bool = True,
        **kwargs,
    ) -> None:
        """Log callback.

        Parameters
        ----------
        model : MuVI
            A MuVI object
        n_epochs : int
            Number of training epochs
        n_checkpoints : int, optional
            Number of times to execute during training,
            by default 10
        active_callbacks : List[str], optional
            List of callback names from LogCallback._CALLBACK_KEYS,
            by default None (all)
        view_wise : bool, optional
            Whether to report scores for each view,
            by default True
        """
        self.kwargs = kwargs
        self.model = model
        self.n_epochs = n_epochs
        self.n_checkpoints = n_checkpoints
        self.view_wise = view_wise

        if active_callbacks is None:
            active_callbacks = self._CALLBACK_KEYS
        self.callback_config = {k: self._default_config[k] for k in active_callbacks}
        self.callback_dict = {
            k: CheckpointCallback(
                self.model,
                self.n_epochs,
                n_checkpoints=v["n_checkpoints"],
                callback=v["callback"],
            )
            for k, v in self.callback_config.items()
        }
        self.threshold = kwargs.get("threshold", 0.1)
        self.should_log = kwargs.get("log", False)

        self.ws = []
        self.zs = []
        self.scores = defaultdict(list)

    @property
    def _default_config(self):
        return {
            "sparsity": {
                "n_checkpoints": self.n_checkpoints,
                "callback": self.sparsity,
            },
            "binary_scores": {
                "n_checkpoints": self.n_checkpoints,
                "callback": self.binary_scores,
            },
            "rmse": {"n_checkpoints": self.n_checkpoints, "callback": self._rmse},
            "r2": {
                "n_checkpoints": self.n_checkpoints,
                "callback": self._variance_explained,
            },
            "log": {
                "n_checkpoints": self.n_checkpoints,
                "callback": self.log,
            },
        }

    @property
    def factor_loadings(self):
        return self.model.get_factor_loadings()

    def get_activations(self, threshold):
        return {vn: np.abs(w) > threshold for vn, w in self.factor_loadings.items()}

    @property
    def factor_scores(self):
        return self.model.get_factor_scores()

    @property
    def masks(self):
        return self.kwargs.get("masks", self.model.get_prior_masks())

    def sparsity(self):
        activations = self.get_activations(self.threshold)
        frac_inactive = {vn: 1 - a.mean() for vn, a in activations.items()}

        str_result = "Average fraction of inactive loadings:\n"
        for vn, fi in frac_inactive.items():
            str_result += f"{vn}: {fi:.3f}, "
            self.scores[f"sparsity_{vn}"].append(fi)
        str_result = str_result[:-2]
        print(str_result)
        return frac_inactive

    def binary_scores(self):
        masks = self.masks
        at = self.kwargs.get("binary_scores_at", 100)

        if masks is None:
            print("Mask is none...")
            return

        thresholds = [self.threshold]
        n_annotated = self.kwargs.get("n_annotated", self.model.n_factors)

        for threshold in thresholds:
            str_result = (
                "Binary scores between prior mask and learned activations "
                + f"with a threshold of {threshold} for top {at} (abs) weights:"
                + "\n`view_name`: (prec, rec, f1)\n"
            )

            informed_views = self.kwargs.get("informed_views", self.model.view_names)
            for vn in informed_views:
                if isinstance(vn, int):
                    vn = self.model.view_names[vn]
                prec, rec, f1, _ = compute_binary_scores_at(
                    masks[vn][:n_annotated, :],
                    self.factor_loadings[vn][:n_annotated, :],
                    threshold=threshold,
                    at=at,
                )
                self.scores[f"precision_{vn}"].append(prec)
                self.scores[f"recall_{vn}"].append(rec)
                self.scores[f"f1_{vn}"].append(f1)
                str_result += f"{vn}: ({prec:.3f}, {rec:.3f}, {f1:.3f})\n"
            str_result = str_result[:-1]
            print(str_result)

    def _rmse(self):
        cov_idx = "all"
        if self.model.n_covariates == 0:
            cov_idx = None
        scores = rmse(
            self.model, cov_idx=cov_idx, factor_wise=False, cov_wise=False, cache=False
        )[0]
        str_result = "RMSE for each view:\n"
        for vn, s in scores.items():
            str_result += f"{vn}: {s:.3f}, "
            self.scores[f"rmse_{vn}"].append(s)
        str_result = str_result[:-2]
        print(str_result)
        return scores

    def _variance_explained(self):
        cov_idx = "all"
        if self.model.n_covariates == 0:
            cov_idx = None
        scores = variance_explained(
            self.model, cov_idx=cov_idx, factor_wise=False, cov_wise=False, cache=False
        )[0]
        str_result = "Variance explained for each view:\n"
        for vn, s in scores.items():
            str_result += f"{vn}: {s:.3f}, "
            self.scores[f"r2_{vn}"].append(s)
        str_result = str_result[:-2]
        print(str_result)
        return scores

    def log(self):
        if self.should_log:
            self.ws.append(self.factor_loadings)
            self.zs.append(self.factor_scores)

    def __call__(self, loss_history):
        for _key, cb in self.callback_dict.items():
            cb(loss_history)

        return False


def compute_binary_scores_at(
    true_mask: np.ndarray,
    learned_w: np.ndarray,
    threshold: float = 0.05,
    at: Optional[int] = None,
):
    """Compute binary scores (precision, recall, F1).

    Parameters
    ----------
    true_mask : np.ndarray
        True underlying mask or the prior mask
    learned_w : np.ndarray
        Factor loadings
    threshold : float, optional
        Threshold to determine active and inactive loadings,
        by default 0.05
    at : int, optional
        Number of the factor loadings to consider,
        sorted by abs value,
        by default None (all)

    Returns
    -------
    precision : float (if average is not None) or array of float, shape =\
        [n_unique_labels]

    recall : float (if average is not None) or array of float, shape =\
        [n_unique_labels]

    fbeta_score : float (if average is not None) or array of float, shape =\
        [n_unique_labels]

    support : None (if average is not None) or array of int, shape =\
        [n_unique_labels]
        The number of occurrences of each label in ``y_true``.

    """
    if at is None:
        at = true_mask.shape[1]

    learned_w_abs = np.abs(learned_w)
    # descending order
    argsort_indices = np.argsort(-learned_w_abs, axis=1)

    sorted_true_mask = np.array(
        list(map(lambda x, y: y[x], argsort_indices, true_mask))
    )
    sorted_learned_mask = (
        np.array(list(map(lambda x, y: y[x], argsort_indices, learned_w_abs)))
        > threshold
    )
    sorted_true_mask = sorted_true_mask[:, :at].flatten()
    sorted_learned_mask = sorted_learned_mask[:, :at].flatten()

    return precision_recall_fscore_support(
        sorted_true_mask, sorted_learned_mask, average="binary"
    )



import logging
import time

from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pandas as pd
import pyro
import pyro.distributions as dist
import torch

from pandas.api.types import is_string_dtype
from pyro.distributions import constraints
from pyro.infer.autoguide.guides import deep_getattr
from pyro.infer.autoguide.guides import deep_setattr
from pyro.nn import PyroModule
from pyro.nn import PyroParam
from pyro.optim import Adam
from pyro.optim import ClippedAdam
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from tqdm import tqdm

from muvi.core.index import _normalize_index


logger = logging.getLogger(__name__)

Index = Union[int, str, List[int], List[str], np.ndarray, pd.Index]
SingleView = Union[np.ndarray, pd.DataFrame]
MultiView = Union[Dict[str, SingleView], List[SingleView]]


class DictDataset(Dataset):
    """Dictionary based PyTorch dataset."""

    def __init__(self, tensor_dict, idx_key="sample_idx", **kwargs):
        self.tensor_dict = tensor_dict
        self.idx_key = idx_key
        # just stores them
        self.kwargs = kwargs
        self._len = tensor_dict[next(iter(tensor_dict))].shape[0]

    def __len__(self):
        return self._len

    def __getitem__(self, index):
        item = {self.idx_key: index}
        for k, v in self.tensor_dict.items():
            item[k] = v[index]
        return item


class MuVI(PyroModule):
    def __init__(
        self,
        observations: MultiView,
        prior_masks: Optional[MultiView] = None,
        covariates: Optional[SingleView] = None,
        prior_confidence: Union[float, str] = 0.99,
        n_factors: Optional[int] = None,
        view_names: Optional[List[str]] = None,
        likelihoods: Optional[Union[Dict[str, str], List[str]]] = None,
        reg_hs: bool = True,
        nmf: Optional[Union[Dict[str, bool], List[bool]]] = None,
        device: str = "cuda",
    ):
        """MuVI module.

        Parameters
        ----------
        observations : MultiView
            Collection of observations as list or dict.
        prior_masks : MultiView, optional
            Prior feature sets as a collection of binary masks,
            by default None
        covariates : SingleView, optional
            Additional observed covariates, by default None
        prior_confidence : float or string
            Confidence of prior belief from 0 to 1 (exclusive),
            typical values are 'low' (0.95), 'med' (0.99) and 'high' (0.999),
            by default 0.99
        n_factors : int, optional
            Number of latent factors,
            can be omitted when providing prior masks,
            by default None
        view_names : List[str], optional
            List of names for each view,
            determines view order as well,
            by default None
        likelihoods : Union[Dict[str, str], List[str]], optional
            Likelihoods for each view,
            either "normal" or "bernoulli",
            by default None (all "normal")
        reg_hs : bool, optional
            Whether to use the regularized version of HS,
            by default True,
            only relevant when NOT using informative priors
        nmf : Union[Dict[str, bool], List[bool]], optional
            Whether to use non-negative matrix factorization,
            by default False
        device : str, optional
            Device to run computations on, by default "cuda" (GPU)
        """
        super().__init__(name="MuVI")
        self.observations = self._setup_observations(observations, view_names)
        self.prior_masks, self.prior_scales = self._setup_prior_masks(
            prior_masks, prior_confidence, n_factors
        )
        self.covariates = self._setup_covariates(covariates)
        self.likelihoods = self._setup_likelihoods(likelihoods)
        self.nmf = self._setup_nmf(nmf)

        self.reg_hs = reg_hs
        self._informed = self.prior_masks is not None
        if not self.reg_hs and self._informed:
            logger.warning(
                "Informative priors require regularized horseshoe, "
                "setting `reg_hs` to True."
            )
            self.reg_hs = True

        self.device = self._setup_device(device)
        self.to(self.device)

        self._model = None
        self._guide = None
        self._built = False
        self._trained = False
        self._training_log: Dict[str, Any] = {}
        self._cache = None

    @property
    def _factor_signs(self):
        signs = 1.0
        # only compute signs if model is trained
        # and there is no nmf constraint
        if self._trained and not any(self.nmf.values()):
            w = self._guide.get_w()
            # TODO: looking at top loadings works better than including all
            # 100 feels a bit arbitrary though
            w = np.array(
                list(map(lambda x, y: y[x], np.argsort(-np.abs(w), axis=1)[:, :100], w))
            )
            signs = (w.sum(axis=1) > 0) * 2 - 1

        return pd.Series(signs, index=self.factor_names, dtype=np.float32)

    def _setup_device(self, device):
        cuda_available = torch.cuda.is_available()

        try:
            mps_available = torch.backends.mps.is_available()
        except AttributeError:
            mps_available = False

        device = str(device).lower()
        if ("cuda" in device and not cuda_available) or (
            device == "mps" and not mps_available
        ):
            logger.warning(f"`{device}` not available...")
            device = "cpu"

        logger.info(f"Running all computations on `{device}`.")
        return torch.device(device)

    def _validate_index(self, idx):
        if not is_string_dtype(idx):
            logger.warning("Transforming to str index.")
            return idx.astype(str)
        return idx

    def _setup_views(self, observations, view_names):
        if view_names is None:
            logger.warning("No view names provided!")
            if isinstance(observations, list):
                logger.info(
                    "Setting the name of each view to `view_idx` for list observations."
                )
                view_names = [f"view_{m}" for m in range(len(observations))]
            if isinstance(observations, dict):
                logger.info(
                    "Setting the view names to the sorted list "
                    "of dictonary keys in observations."
                )
                view_names = sorted(observations.keys())

        # if list convert to dict
        if isinstance(observations, list):
            observations = dict(zip(view_names, observations))

        _view_names = []
        for vn in view_names:
            if vn not in observations:
                logger.warning(f"View `{vn}` not found in observations, skipping.")
            else:
                _view_names.append(vn)
        view_names = _view_names
        if len(view_names) == 0:
            raise ValueError(
                "No views found, check if `view_names` matches the dictionary keys!"
            )

        if len(view_names) < len(observations):
            logger.warning(
                "Number of views is larger than the length of `view_names`, "
                "using only the subset of views defined in `view_names`."
            )

        observations = {vn: observations[vn] for vn in view_names}

        n_views = len(observations)
        if n_views == 1:
            logger.warning("Running MuVI on a single view.")

        view_names = pd.Index(view_names)

        self.n_views = n_views
        self.view_names = self._validate_index(view_names)
        return observations

    def _setup_samples(self, observations):
        n_samples = 0
        sample_names = None

        for vn in self.view_names:
            view_obs = observations[vn]
            view_n_samples = view_obs.shape[0]

            if n_samples == 0:
                n_samples = view_n_samples

            if n_samples != view_n_samples:
                raise ValueError(
                    f"View `{vn}` has {view_n_samples} samples "
                    f"instead of {n_samples}, "
                    "all views must have the same number of samples."
                )

            if isinstance(view_obs, pd.DataFrame):
                logger.info("pd.DataFrame detected.")

                view_sample_names = view_obs.index.copy()

                if sample_names is None:
                    logger.info(
                        f"Storing the index of the view `{vn}` as sample names."
                    )
                    sample_names = view_sample_names

                if any(sample_names != view_sample_names):
                    logger.info(
                        f"Sample names for view `{vn}` do not match "
                        f"the sample names of view `{self.view_names[0]}`, "
                        f"sorting names according to view `{self.view_names[0]}`."
                    )
                    view_obs = view_obs.loc[sample_names, :]

        if sample_names is None:
            logger.info("Setting the name of each sample to `sample_idx`.")
            sample_names = pd.Index([f"sample_{i}" for i in range(n_samples)])

        self.sample_names = self._validate_index(sample_names)
        self.n_samples = n_samples

        return observations

    def _setup_features(self, observations):
        n_features = {vn: 0 for vn in self.view_names}
        feature_names = {vn: None for vn in self.view_names}

        for vn in self.view_names:
            view_obs = observations[vn]
            n_features[vn] = view_obs.shape[1]
            if isinstance(view_obs, pd.DataFrame):
                logger.info("pd.DataFrame detected.")
                feature_names[vn] = view_obs.columns.copy()
            if feature_names[vn] is None:
                logger.info(
                    f"Setting the name of each feature in `{vn}` to `{vn}_feature_idx`."
                )
                feature_names[vn] = pd.Index(
                    [f"{vn}_feature_{j}" for j in range(n_features[vn])]
                )

            feature_names[vn] = self._validate_index(feature_names[vn])

        self.feature_names = feature_names
        self.n_features = n_features

        return observations

    def _setup_observations(self, observations, view_names):
        if observations is None:
            raise ValueError(
                "Observations is None, please provide a valid list of observations."
            )

        observations = self._setup_views(observations, view_names)
        # from now on only working with nested dictonaries
        observations = self._setup_samples(observations)
        observations = self._setup_features(observations)

        # keep only numpy arrays, convert to np.float32 dtypes
        return {
            vn: (
                obs.to_numpy(dtype=np.float32)
                if isinstance(obs, pd.DataFrame)
                else np.array(obs, dtype=np.float32)
            )
            for vn, obs in observations.items()
        }

    def _setup_prior_masks(self, masks, confidence, n_factors):
        informed = masks is not None and len(masks) > 0
        valid_n_factors = n_factors is not None and n_factors > 0
        if not informed and not valid_n_factors:
            raise ValueError(
                "Invalid latent configuration, "
                "please provide either a collection of prior masks, "
                "or set `n_factors` to a positive integer."
            )

        if not informed:
            self.n_factors = n_factors
            # TODO: duplicate line...see below
            self.factor_names = pd.Index([f"factor_{k}" for k in range(n_factors)])
            return None, None

        if confidence is None:
            logger.info(
                "No prior confidence provided, setting `prior_confidence` to 0.99."
            )
            confidence = 0.99

        if isinstance(confidence, str):
            try:
                confidence = {"low": 0.95, "med": 0.99, "high": 0.999}[
                    confidence.lower()
                ]
            except KeyError as e:
                logger.error(e)
                raise

        if not (0 < confidence < 1.0):
            raise ValueError(
                "Invalid prior confidence, please provide a positive value less than 1."
            )

        # if list convert to dict
        if isinstance(masks, list):
            masks = {self.view_names[m]: mask for m, mask in enumerate(masks)}

        informed_views = [vn for vn in self.view_names if vn in masks]

        n_prior_factors = masks[informed_views[0]].shape[0]
        if n_factors is None:
            n_factors = n_prior_factors

        if n_prior_factors > n_factors:
            logger.warning(
                f"Prior mask informs more factors ({n_prior_factors}) "
                f"than the pre-defined `n_factors` ({n_factors}). "
                f"Updating `n_factors` to {n_prior_factors}."
            )
            n_factors = n_prior_factors

        n_dense_factors = 0
        if n_prior_factors < n_factors:
            logger.warning(
                f"Prior mask informs fewer factors ({n_prior_factors}) "
                f"than the pre-defined `n_factors` ({n_factors}). "
                f"Informing only the first {n_prior_factors} factors, "
                "the rest remains uninformed."
            )
            # extend all prior masks with additional uninformed factors
            n_dense_factors = n_factors - n_prior_factors

        factor_names = None
        for vn in self.view_names:
            # fill uninformed views
            # will only execute if a subset of views are being informed
            if vn not in masks:
                logger.info(
                    f"Mask for view `{vn}` not found, assuming `{vn}` to be uninformed."
                )
                masks[vn] = np.zeros((n_prior_factors, self.n_features[vn])).astype(
                    bool
                )
                continue
            view_mask = masks[vn]
            if view_mask.shape[0] != n_prior_factors:
                raise ValueError(
                    f"Mask `{vn}` has {view_mask.shape[0]} factors "
                    f"instead of {n_prior_factors}, "
                    "all masks must have the same number of factors."
                )
            if view_mask.shape[1] != self.n_features[vn]:
                raise ValueError(
                    f"Mask `{vn}` has {view_mask.shape[1]} features "
                    f"instead of {self.n_features[vn]}, "
                    "each mask must match the number of features of its view."
                )

            if isinstance(view_mask, pd.DataFrame):
                logger.info("pd.DataFrame detected.")

                mask_factor_names = view_mask.index.copy()
                if factor_names is None:
                    logger.info(
                        f"Storing the index of the mask `{vn}` as factor names."
                    )
                    factor_names = mask_factor_names
                if any(factor_names != mask_factor_names):
                    logger.info(
                        f"Factor names for mask `{vn}` "
                        f"do not match the factor names of mask `{informed_views[0]}`, "
                        f"sorting names according to mask `{informed_views[0]}`."
                    )
                    masks[vn] = view_mask.loc[factor_names, :]

                if any(self.feature_names[vn] != view_mask.columns):
                    logger.info(
                        f"Feature names for mask `{vn}` "
                        "do not match the feature names of its corresponding view, "
                        "sorting names according to the view features."
                    )
                    masks[vn] = view_mask.loc[:, self.feature_names[vn]]

        if factor_names is None:
            factor_names = [f"factor_{k}" for k in range(n_prior_factors)]
        if n_dense_factors > 0:
            factor_names = list(factor_names) + [
                f"dense_{k}" for k in range(n_dense_factors)
            ]
        self.factor_names = self._validate_index(pd.Index(factor_names))

        # keep only numpy arrays
        prior_masks = {
            vn: (
                vm.to_numpy().astype(bool)
                if isinstance(vm, pd.DataFrame)
                else vm.astype(bool)
            )
            for vn, vm in masks.items()
        }
        # add dense factors if necessary
        if n_dense_factors > 0:
            prior_masks = {
                vn: np.concatenate(
                    [vm, np.ones((n_dense_factors, self.n_features[vn])).astype(bool)]
                )
                for vn, vm in masks.items()
            }

        prior_scales = {
            vn: np.clip(vm.astype(np.float32) + (1.0 - confidence), 1e-4, 1.0)
            for vn, vm in prior_masks.items()
        }
        self.n_factors = n_factors
        return prior_masks, prior_scales

    def _setup_likelihoods(self, likelihoods):
        if likelihoods is None:
            likelihoods = ["normal" for _ in range(self.n_views)]
        if isinstance(likelihoods, list):
            likelihoods = {self.view_names[m]: ll for m, ll in enumerate(likelihoods)}
        likelihoods = {vn: likelihoods.get(vn, "normal") for vn in self.view_names}
        logger.info(f"Likelihoods set to `{likelihoods}`.")
        return likelihoods

    def _setup_nmf(self, nmf):
        if nmf is None:
            nmf = [False for _ in range(self.n_views)]
        if isinstance(nmf, bool):
            nmf = [nmf for _ in range(self.n_views)]
        if isinstance(nmf, list):
            nmf = {self.view_names[m]: _nmf for m, _nmf in enumerate(nmf)}
        nmf = {vn: nmf.get(vn, False) for vn in self.view_names}
        logger.info(f"NMF set to `{nmf}`.")
        return nmf

    def _setup_covariates(self, covariates):
        n_covariates = 0
        if covariates is not None:
            n_covariates = covariates.shape[1]

        covariate_names = None
        if isinstance(covariates, pd.DataFrame):
            logger.info("pd.DataFrame detected.")
            if self.n_samples != covariates.shape[0]:
                raise ValueError(
                    f"Number of observed samples for ({self.n_samples}) "
                    "does not match the number of samples "
                    f"for the covariates ({covariates.shape[0]})."
                )

            if any(self.sample_names != covariates.index):
                logger.info(
                    "Sample names for the covariates "
                    "do not match the sample names of the observations, "
                    "sorting names according to the observations."
                )
                covariates = covariates.loc[self.sample_names, :]
            covariate_names = covariates.columns.copy()
            covariates = covariates.to_numpy()
        if covariate_names is None:
            covariate_names = pd.Index([f"covariate_{k}" for k in range(n_covariates)])

        self.n_covariates = n_covariates
        self.covariate_names = covariate_names
        return covariates

    def _raise_untrained_error(self):
        if not self._trained:
            raise AttributeError(
                "Requested attribute cannot be found on an untrained model! "
                "Run the `fit` method to train a MuVI model."
            )
        return True

    def _get_view_attr(
        self,
        attr,
        view_idx,
        feature_idx,
        other_idx,
        other_names,
        as_df,
    ):
        if view_idx is None and isinstance(feature_idx, dict):
            view_idx = []
            for key in feature_idx:
                vi = key
                if isinstance(vi, int):
                    vi = self.view_names[vi]
                view_idx.append(vi)
        if view_idx is None:
            raise IndexError(
                "Invalid indices, `view_idx` is None "
                "and `feature_idx` is not a dictionary."
            )
        # get relevant view names
        view_names = _normalize_index(view_idx, self.view_names, as_idx=False)
        if isinstance(feature_idx, (str, int)):
            if feature_idx == "all":
                feature_idx = ["all" for _ in range(len(view_names))]
            else:
                feature_idx = [feature_idx]

        # convert to list if any list-like indexing
        if isinstance(feature_idx, (list, np.ndarray, pd.Index)):
            feature_idx = list(feature_idx)

        # check if valid combination of indices
        if len(view_names) == 1 and len(feature_idx) != 1:
            if isinstance(feature_idx[0], list):
                logger.warning(
                    "`feature_idx` suggests indices for more than one view, "
                    "keeping only the first list of indices."
                )
                feature_idx = feature_idx[0]
            feature_idx = [feature_idx]

        if len(view_names) != len(feature_idx):
            logger.warning(
                "`view_idx` does not match the keys of `feature_idx`, "
                "`view_idx` has precedence over `feature_idx`."
            )

        if isinstance(feature_idx, list):
            feature_idx = {vn: feature_idx[m] for m, vn in enumerate(view_names)}

        if isinstance(feature_idx, dict):
            feature_idx = {
                (self.view_names[k] if isinstance(k, int) else k): v
                for k, v in feature_idx.items()
            }

        # normalise dictionary
        feature_idx = {
            vn: _normalize_index(feature_idx[vn], self.feature_names[vn])
            for vn in view_names
        }
        other_idx = _normalize_index(other_idx, other_names)

        attr = {vn: attr[vn][other_idx, :][:, feature_idx[vn]] for vn in view_names}
        if as_df:
            attr = {
                vn: pd.DataFrame(
                    attr[vn],
                    index=other_names[other_idx],
                    columns=self.feature_names[vn][feature_idx[vn]],
                )
                for vn in view_names
            }
        return attr

    def _get_shared_attr(
        self,
        attr,
        shared_idx,
        shared_names,
        other_idx,
        other_names,
        as_df,
    ):
        shared_idx = _normalize_index(shared_idx, shared_names)
        other_idx = _normalize_index(other_idx, other_names)
        attr = attr[shared_idx, :][:, other_idx]
        if as_df:
            attr = pd.DataFrame(
                attr,
                index=shared_names[shared_idx],
                columns=other_names[other_idx],
            )
        return attr

    def _get_sample_attr(
        self,
        attr,
        sample_idx,
        other_idx,
        other_names,
        as_df,
    ):
        return self._get_shared_attr(
            attr, sample_idx, self.sample_names, other_idx, other_names, as_df
        )

    def get_observations(
        self,
        view_idx: Index = "all",
        sample_idx: Index = "all",
        feature_idx: Union[Index, List[Index], Dict[str, Index]] = "all",
        as_df: bool = False,
    ):
        """Get observations.

        Parameters
        ----------
        view_idx : Index, optional
            View index, by default "all"
        sample_idx : Index, optional
            Sample index, by default "all"
        feature_idx : Union[Index, List[Index], Dict[str, Index]], optional
            Feature index, by default "all"
        as_df : bool, optional
            Whether to return a pandas dataframe,
            by default False

        Returns
        -------
        dict
            Dictionary with view names as keys,
            and np.ndarray or pd.DataFrame as values.
        """
        return self._get_view_attr(
            self.observations,
            view_idx,
            feature_idx,
            other_idx=sample_idx,
            other_names=self.sample_names,
            as_df=as_df,
        )

    def _mode(self, param, likelihood):
        if likelihood == "bernoulli":
            return (param > 0.5).astype(np.int32)
        return param

    def get_reconstructed(
        self,
        view_idx: Index = "all",
        sample_idx: Index = "all",
        feature_idx: Union[Index, List[Index], Dict[str, Index]] = "all",
        factor_idx: Index = "all",
        cov_idx: Index = "all",
        as_df: bool = False,
    ):
        """Get reconstructed observations.

        Parameters
        ----------
        view_idx : Index, optional
            View index, by default "all"
        sample_idx : Index, optional
            Sample index, by default "all"
        feature_idx : Union[Index, List[Index], Dict[str, Index]], optional
            Feature index, by default "all"
        factor_idx : Index, optional
            Factor index, by default "all"
        cov_idx : Index, optional
            Covariate index, by default "all"
        as_df : bool, optional
            Whether to return a pandas dataframe,
            by default False

        Returns
        -------
        dict
            Dictionary with view names as keys,
            and np.ndarray or pd.DataFrame as values.
        """
        obs_hat = {
            vn: np.zeros((self.n_samples, self.n_features[vn]))
            for vn in _normalize_index(view_idx, self.view_names, as_idx=False)
        }

        if self.n_factors > 0:
            factor_scores = self.get_factor_scores(sample_idx, factor_idx)
            factor_loadings = self.get_factor_loadings(
                view_idx, factor_idx, feature_idx
            )
            obs_hat = {
                vn: obs + (factor_scores @ factor_loadings[vn])
                for vn, obs in obs_hat.items()
            }

        if self.n_covariates > 0:
            covariates = self.get_covariates(sample_idx, cov_idx)
            cov_coefficients = self.get_covariate_coefficients(
                view_idx, cov_idx, feature_idx
            )
            obs_hat = {
                vn: obs + (covariates @ cov_coefficients[vn])
                for vn, obs in obs_hat.items()
            }

        for vn in obs_hat:
            obs_hat[vn] = self._mode(obs_hat[vn], self.likelihoods[vn])

        if as_df:
            obs_hat = {
                vn: pd.DataFrame(
                    obs, index=self.sample_names, columns=self.feature_names[vn]
                )
                for vn, obs in obs_hat.items()
            }

        return obs_hat

    def get_imputed(
        self,
        view_idx: Index = "all",
        sample_idx: Index = "all",
        feature_idx: Union[Index, List[Index], Dict[str, Index]] = "all",
        as_df: bool = False,
    ):
        """Get imputed observations.

        Parameters
        ----------
        view_idx : Index, optional
            View index, by default "all"
        sample_idx : Index, optional
            Sample index, by default "all"
        feature_idx : Union[Index, List[Index], Dict[str, Index]], optional
            Feature index, by default "all"
        as_df : bool, optional
            Whether to return a pandas dataframe,
            by default False

        Returns
        -------
        dict
            Dictionary with view names as keys,
            and np.ndarray or pd.DataFrame as values.
        """

        obs = self.get_observations(view_idx, sample_idx, feature_idx, as_df=True)
        obs_hat = self.get_reconstructed(view_idx, sample_idx, feature_idx, as_df=True)
        obs_imputed = {vn: obs[vn].fillna(obs_hat[vn]) for vn in obs}

        if not as_df:
            obs_imputed = {
                vn: obs_imp.to_numpy() for vn, obs_imp in obs_imputed.items()
            }
        return obs_imputed

    def get_prior_masks(
        self,
        view_idx: Index = "all",
        factor_idx: Index = "all",
        feature_idx: Union[Index, List[Index], Dict[str, Index]] = "all",
        as_df: bool = False,
    ):
        """Get prior masks.

        Parameters
        ----------
        view_idx : Index, optional
            View index, by default "all"
        factor_idx : Index, optional
            Factor index, by default "all"
        feature_idx : Union[Index, List[Index], Dict[str, Index]], optional
            Feature index, by default "all"
        as_df : bool, optional
            Whether to return a pandas dataframe,
            by default False

        Returns
        -------
        dict
            Dictionary with view names as keys,
            and np.ndarray or pd.DataFrame as values.
        """

        return self._get_view_attr(
            self.prior_masks,
            view_idx,
            feature_idx,
            other_idx=factor_idx,
            other_names=self.factor_names,
            as_df=as_df,
        )

    def get_factor_loadings(
        self,
        view_idx: Index = "all",
        factor_idx: Index = "all",
        feature_idx: Union[Index, List[Index], Dict[str, Index]] = "all",
        as_df: bool = False,
    ):
        """Get factor loadings.

        Parameters
        ----------
        view_idx : Index, optional
            View index, by default "all"
        factor_idx : Index, optional
            Factor index, by default "all"
        feature_idx : Union[Index, List[Index], Dict[str, Index]], optional
            Feature index, by default "all"
        as_df : bool, optional
            Whether to return a pandas dataframe,
            by default False

        Returns
        -------
        dict
            Dictionary with view names as keys,
            and np.ndarray or pd.DataFrame as values.
        """
        self._raise_untrained_error()

        ws = self._guide.get_w(as_list=True)
        ws = [w * self._factor_signs.to_numpy()[:, np.newaxis] for w in ws]
        return self._get_view_attr(
            {vn: ws[m] for m, vn in enumerate(self.view_names)},
            view_idx,
            feature_idx,
            other_idx=factor_idx,
            other_names=self.factor_names,
            as_df=as_df,
        )

    def get_covariate_coefficients(
        self,
        view_idx: Index = "all",
        cov_idx: Index = "all",
        feature_idx: Union[Index, List[Index], Dict[str, Index]] = "all",
        as_df: bool = False,
    ):
        """Get factor loadings.

        Parameters
        ----------
        view_idx : Index, optional
            View index, by default "all"
        cov_idx : Index, optional
            Covariate index, by default "all"
        feature_idx : Union[Index, List[Index], Dict[str, Index]], optional
            Feature index, by default "all"
        as_df : bool, optional
            Whether to return a pandas dataframe,
            by default False

        Returns
        -------
        dict
            Dictionary with view names as keys,
            and np.ndarray or pd.DataFrame as values.
        """
        self._raise_untrained_error()

        betas = self._guide.get_beta(as_list=True)
        return self._get_view_attr(
            {vn: betas[m] for m, vn in enumerate(self.view_names)},
            view_idx,
            feature_idx,
            other_idx=cov_idx,
            other_names=self.covariate_names,
            as_df=as_df,
        )

    def get_factor_scores(
        self, sample_idx: Index = "all", factor_idx: Index = "all", as_df: bool = False
    ):
        """Get factor scores.

        Parameters
        ----------
        sample_idx : Index, optional
            Sample index, by default "all"
        factor_idx : Index, optional
            Factor index, by default "all"
        as_df : bool, optional
            Whether to return a pandas dataframe,
            by default False

        Returns
        -------
        Union[np.ndarray, pd.DataFrame]
            A single np.ndarray or pd.DataFrame of shape `n_samples` x `n_factors`.
        """
        self._raise_untrained_error()

        return self._get_sample_attr(
            self._guide.get_z() * self._factor_signs.to_numpy(),
            sample_idx,
            other_idx=factor_idx,
            other_names=self.factor_names,
            as_df=as_df,
        )

    def get_covariates(
        self, sample_idx: Index = "all", cov_idx: Index = "all", as_df: bool = False
    ):
        """Get factor scores.

        Parameters
        ----------
        sample_idx : Index, optional
            Sample index, by default "all"
        cov_idx : Index, optional
            Covariate index, by default "all"
        as_df : bool, optional
            Whether to return a pandas dataframe,
            by default False

        Returns
        -------
        Union[np.ndarray, pd.DataFrame]
            A single np.ndarray or pd.DataFrame of shape `n_samples` x `n_covariates`.
        """
        return self._get_sample_attr(
            self.covariates,
            sample_idx,
            other_idx=cov_idx,
            other_names=self.covariate_names,
            as_df=as_df,
        )

    def _setup_model_guide(self, batch_size: int, scale_elbo: bool):
        """Setup model and guide.

        Parameters
        ----------
        batch_size : int
            Batch size when subsampling

        Returns
        -------
        bool
            Whether the build was successful
        """
        if not self._built:
            if not self._informed:
                logger.warning(
                    "No prior feature sets provided, running model uninformed."
                )

            self._model = MuVIModel(
                self.n_samples,
                n_subsamples=batch_size,
                n_features=[self.n_features[vn] for vn in self.view_names],
                n_factors=self.n_factors,
                n_covariates=self.n_covariates,
                likelihoods=[self.likelihoods[vn] for vn in self.view_names],
                reg_hs=self.reg_hs,
                nmf=[self.nmf[vn] for vn in self.view_names],
                scale_elbo=scale_elbo,
                device=self.device,
            )
            self._guide = MuVIGuide(self._model)
            self._built = True
        return self._built

    def _setup_optimizer(
        self, batch_size: int, n_epochs: int, learning_rate: float, optimizer: str
    ):
        """Setup SVI optimizer.

        Parameters
        ----------
        batch_size : int
            Batch size
        n_epochs : int
            Number of epochs, needed to schedule learning rate decay
        learning_rate : float
            Learning rate
        optimizer : str
            Optimizer as string, 'adam' or 'clipped'

        Returns
        -------
        pyro.optim.PyroOptim
            pyro or torch optimizer object
        """

        optim = Adam({"lr": learning_rate, "betas": (0.95, 0.999)})
        if optimizer.lower() == "clipped":
            n_iterations = int(n_epochs * (self.n_samples // batch_size))
            logger.info(f"Decaying learning rate over {n_iterations} iterations.")
            gamma = 0.1
            lrd = gamma ** (1 / n_iterations)
            optim = ClippedAdam({"lr": learning_rate, "lrd": lrd})

        self._optimizer = optim
        return self._optimizer

    def _setup_svi(
        self,
        optimizer: pyro.optim.PyroOptim,
        n_particles: int,
        scale: bool = True,
    ):
        """Setup stochastic variational inference.

        Parameters
        ----------
        optimizer : pyro.optim.PyroOptim
            pyro or torch optimizer
        n_particles : int
            Number of particles/samples used to form the ELBO (gradient) estimators
        scale : bool, optional
            Whether to scale ELBO by the number of samples, by default True

        Returns
        -------
        pyro.infer.SVI
            pyro SVI object
        """
        scaler = 1.0
        if scale:
            scaler = 1.0 / self.n_samples

        loss = pyro.infer.TraceMeanField_ELBO(
            retain_graph=True,
            num_particles=n_particles,
            vectorize_particles=True,
        )

        svi = pyro.infer.SVI(
            model=pyro.poutine.scale(self._model, scale=scaler),
            guide=pyro.poutine.scale(self._guide, scale=scaler),
            optim=optimizer,
            loss=loss,
        )

        self._svi = svi
        return self._svi

    def _setup_training_data(self):
        """Setup training components.
        Convert observations, covariates and prior scales to torch.Tensor and
        extract mask of missing values to mask SVI updates.

        Returns
        -------
        dict
            Dictionary of training data,
            same keys are expected as in the forward method of the MuVIModel
        """
        train_obs = torch.cat(
            [torch.Tensor(self.observations[vn]) for vn in self.view_names], 1
        )
        mask_obs = ~torch.isnan(train_obs)
        # replace all nans with zeros
        # self.presence mask takes care of gradient updates
        train_obs = torch.nan_to_num(train_obs)

        train_covs = None
        if self.covariates is not None:
            train_covs = torch.Tensor(self.covariates)

        train_prior_scales = None
        if self._informed:
            train_prior_scales = torch.cat(
                [torch.Tensor(self.prior_scales[vn]) for vn in self.view_names], 1
            )

        # make sure keys match the arguments in the forward method of the MuVIModel
        training_data = {
            "obs": train_obs,
            "mask": mask_obs,
            "covs": train_covs,
            "prior_scales": train_prior_scales,
        }
        return {k: v for k, v in training_data.items() if v is not None}

    def fit(
        self,
        batch_size: int = 0,
        n_epochs: int = 1000,
        n_particles: int = 0,
        learning_rate: float = 0.005,
        optimizer: str = "clipped",
        scale_elbo: bool = True,
        callbacks: Optional[List[Callable]] = None,
        verbose: bool = True,
        seed: Optional[int] = None,
    ):
        """Perform inference.

        Parameters
        ----------
        batch_size : int, optional
            Batch size, by default 0 (all samples)
        n_epochs : int, optional
            Number of iterations over the whole dataset,
            by default 1000
        n_particles : int, optional
            Number of particles/samples used to form the ELBO (gradient) estimators,
            by default 0 (1000 // batch_size)
        learning_rate : float, optional
            Learning rate, by default 0.005
        scale_elbo : bool
            Whether to scale the ELBO across views, by default True
        optimizer : str, optional
            Optimizer as string, 'adam' or 'clipped', by default "clipped"
        callbacks : List[Callable], optional
            List of callbacks during training, by default None
        verbose : bool
            Whether to log progress, by default True
        seed : int, optional
            Training seed, by default None
        """

        # if invalid or out of bounds set to n_samples
        if batch_size is None or not (0 < batch_size <= self.n_samples):
            batch_size = self.n_samples

        if n_particles < 1:
            n_particles = max(1, 1000 // batch_size)
        if n_particles > 1:
            logger.info(f"Using {n_particles} particles in parallel.")
        logger.info("Preparing model and guide...")
        self._setup_model_guide(batch_size, scale_elbo)
        logger.info("Preparing optimizer...")
        opt = self._setup_optimizer(batch_size, n_epochs, learning_rate, optimizer)
        logger.info("Preparing SVI...")
        svi = self._setup_svi(opt, n_particles, scale=True)
        logger.info("Preparing training data...")
        training_data = self._setup_training_data()
        training_prior_scales = training_data.pop("prior_scales", None)
        if training_prior_scales is not None:
            training_prior_scales = training_prior_scales.to(self.device)

        if batch_size < self.n_samples:
            logger.info(f"Using batches of size `{batch_size}`.")

            data_loader = DataLoader(
                DictDataset(training_data, idx_key="sample_idx"),
                batch_size=batch_size,
                shuffle=True,
                num_workers=1,
                pin_memory=str(self.device) != "cpu",
                drop_last=False,
            )

            def _step():
                iteration_loss = 0
                for _, tensor_dict in enumerate(data_loader):
                    iteration_loss += svi.step(
                        **{k: v.to(self.device) for k, v in tensor_dict.items()},
                        prior_scales=training_prior_scales,
                    )
                return iteration_loss

        else:
            logger.info("Using complete dataset.")
            # move all data to device once
            training_data = {k: v.to(self.device) for k, v in training_data.items()}

            def _step():
                return svi.step(
                    None,
                    **training_data,
                    prior_scales=training_prior_scales,
                )

        if seed is not None:
            try:
                seed = int(seed)
            except ValueError:
                logger.warning(f"Could not convert `{seed}` to integer.")
                seed = None

        if seed is None:
            seed = int(time.strftime("%y%m%d%H%M"))

        logger.info(f"Setting training seed to `{seed}`.")
        pyro.set_rng_seed(seed)
        # clean start
        logger.info("Cleaning parameter store.")
        pyro.enable_validation(True)
        pyro.clear_param_store()

        logger.info("Starting training...")
        self._trained = True
        stop_early = False
        history = []
        pbar = range(n_epochs)
        if verbose:
            pbar = tqdm(pbar)
            window_size = 5

        try:
            for epoch_idx in pbar:
                epoch_loss = _step()
                history.append(epoch_loss)
                if verbose and (
                    epoch_idx % window_size == 0 or epoch_idx == n_epochs - 1
                ):
                    pbar.set_postfix({"ELBO": epoch_loss})
                if callbacks is not None:
                    # TODO: dont really like this, a bit sloppy
                    stop_early = any(callback(history) for callback in callbacks)
                    if stop_early:
                        break
        except KeyboardInterrupt:
            logger.info("Keyboard interrupt, stopping training and saving progress...")

        self._training_log = {
            "n_epochs": n_epochs,
            "batch_size": batch_size,
            "n_particles": n_particles,
            "learning_rate": learning_rate,
            "optimizer": optimizer,
            "verbose": verbose,
            "seed": seed,
            "history": history,
            "n_iter": len(history),
            "stop_early": stop_early,
        }
        logger.info("Call `model._training_log` to inspect the training progress.")
        # reset cache in case it was initialized by any of the callbacks
        self._cache = None


class MuVIModel(PyroModule):
    def __init__(
        self,
        n_samples: int,
        n_subsamples: int,
        n_features: List[int],
        n_factors: int,
        n_covariates: int,
        likelihoods: List[str],
        global_prior_scale: float = 1.0,
        reg_hs: bool = True,
        nmf: Optional[List[bool]] = None,
        scale_elbo: bool = True,
        device=None,
    ):
        """MuVI generative model.

        Parameters
        ----------
        n_samples : int
            Number of samples
        n_subsamples : int
            Number of subsamples (batch size)
        n_features : List[int]
            Number of features as list for each view
        n_factors : int
            Number of latent factors
        n_covariates : int
            Number of covariates
        likelihoods : List[str], optional
            List of likelihoods for each view,
            either "normal" or "bernoulli", by default None
        global_prior_scale : float, optional
            Determine the level of global sparsity, by default 1.0
        reg_hs : bool, optional
            Whether to use the regularized version of HS,
            by default True
        nmf : List[bool], optional
            Whether to use non-negative matrix factorization,
            by default empty (False for all views)
        scale_elbo : bool
            Whether to scale the ELBO across views, by default True
        device : str, optional
            Device to run computations on, by default "cuda" (GPU)
        """
        super().__init__(name="MuVIModel")
        self.n_samples = n_samples
        self.n_subsamples = n_subsamples
        self.n_features = n_features
        self.feature_offsets = [0, *np.cumsum(self.n_features).tolist()]
        self.n_views = len(self.n_features)
        self.n_factors = n_factors
        self.n_covariates = n_covariates
        self.likelihoods = likelihoods
        self.same_likelihood = len(set(self.likelihoods)) == 1
        self.global_prior_scale = (
            1.0 if global_prior_scale is None else global_prior_scale
        )
        self.reg_hs = reg_hs
        if nmf is None:
            nmf = [False for _ in range(self.n_views)]
        self.nmf = nmf
        # only needed if nmf is True
        self.pos_transform = torch.nn.ReLU()

        self.scale_elbo = scale_elbo
        self.view_scales = np.ones(self.n_views)
        if self.scale_elbo and self.n_views > 1:
            self.view_scales = (self.n_views / (self.n_views - 1)) * (
                1.0 - np.array([nf / sum(n_features) for nf in n_features])
            )
        self.device = device

    def get_plate(self, name: str, **kwargs):
        """Get the sampling plate.

        Parameters
        ----------
        name : str
            Name of the plate

        Returns
        -------
        PlateMessenger
            A pyro plate.
        """
        plate_kwargs = {
            "view": {"name": "view", "size": self.n_views, "dim": -1},
            "factor_left": {"name": "factor_left", "size": self.n_factors, "dim": -2},
            "sample": {"name": "sample", "size": self.n_samples, "dim": -2},
            "covariate": {"name": "covariate", "size": self.n_covariates, "dim": -2},
        }
        for m, n_features in zip(range(self.n_views), self.n_features):
            plate_kwargs[f"feature_{m}"] = {
                "name": f"feature_{m}",
                "size": n_features,
                "dim": -1,
            }
        return pyro.plate(device=self.device, **{**plate_kwargs[name], **kwargs})

    def _zeros(self, size):
        return torch.zeros(size, device=self.device)

    def _ones(self, size):
        return torch.ones(size, device=self.device)

    def forward(
        self,
        sample_idx: torch.Tensor,
        obs: torch.Tensor,
        mask: torch.Tensor,
        covs: Optional[torch.Tensor] = None,
        prior_scales: Optional[torch.Tensor] = None,
    ):
        """Generate samples.

        Parameters
        ----------
        obs : torch.Tensor
            Observations to condition the model on
        mask : torch.Tensor
            Binary mask of missing data
        covs : torch.Tensor, optional
            Additional covariate matrix, by default None
        prior_scales : torch.Tensor, optional
            Local prior scales with prior information,
            by default None

        Returns
        -------
        dict
            Samples from each sampling site
        """

        output_dict = {}

        view_plate = self.get_plate("view")
        factor_l_plate = self.get_plate("factor_left")
        feature_plates = [self.get_plate(f"feature_{m}") for m in range(self.n_views)]
        sample_plate = self.get_plate("sample", subsample=sample_idx)

        if self.n_covariates > 0:
            covariate_plate = self.get_plate("covariate")

        with view_plate:
            output_dict["view_scale"] = pyro.sample(
                "view_scale", dist.HalfCauchy(self._ones((1,)))
            )
            with factor_l_plate:
                output_dict["factor_scale"] = pyro.sample(
                    "factor_scale",
                    dist.HalfCauchy(self._ones((1,))),
                )

        for m in range(self.n_views):
            with feature_plates[m]:
                with factor_l_plate:
                    output_dict[f"local_scale_{m}"] = pyro.sample(
                        f"local_scale_{m}",
                        dist.HalfCauchy(self._ones((1,))),
                    )

                    w_scale = (
                        output_dict[f"local_scale_{m}"]
                        * output_dict["factor_scale"][..., m : m + 1]
                        * output_dict["view_scale"][..., m : m + 1]
                    )

                    if self.reg_hs:
                        output_dict[f"caux_{m}"] = pyro.sample(
                            f"caux_{m}",
                            dist.InverseGamma(
                                0.5 * self._ones((1,)), 0.5 * self._ones((1,))
                            ),
                        )
                        c = torch.sqrt(output_dict[f"caux_{m}"])
                        if prior_scales is not None:
                            c = (
                                c
                                * prior_scales[
                                    :,
                                    self.feature_offsets[m] : self.feature_offsets[
                                        m + 1
                                    ],
                                ]
                            )
                        w_scale = (self.global_prior_scale * c * w_scale) / torch.sqrt(
                            c**2 + w_scale**2
                        )

                    output_dict[f"w_{m}"] = pyro.sample(
                        f"w_{m}",
                        dist.Normal(
                            self._zeros((1,)),
                            w_scale,
                        ),
                    )

                    if self.nmf[m]:
                        output_dict[f"w_{m}"] = self.pos_transform(
                            output_dict[f"w_{m}"]
                        )

                if self.n_covariates > 0:
                    with covariate_plate:
                        output_dict[f"beta_{m}"] = pyro.sample(
                            f"beta_{m}",
                            dist.Normal(self._zeros((1,)), self._ones((1,))),
                        )

                output_dict[f"sigma_{m}"] = pyro.sample(
                    f"sigma_{m}",
                    dist.InverseGamma(self._ones((1,)), self._ones((1,))),
                )

        with sample_plate:
            output_dict["z"] = pyro.sample(
                "z",
                dist.Normal(
                    self._zeros((self.n_factors,)),
                    self._ones((self.n_factors,)),
                ),
            )
            if any(self.nmf):
                output_dict["z"] = self.pos_transform(output_dict["z"])
            for m in range(self.n_views):
                with feature_plates[m]:
                    y_loc = torch.matmul(output_dict["z"], output_dict[f"w_{m}"])
                    if self.n_covariates > 0:
                        y_loc = y_loc + torch.matmul(covs, output_dict[f"beta_{m}"])

                    if self.likelihoods[m] == "normal":
                        y_dist = dist.Normal(
                            y_loc,
                            torch.sqrt(output_dict[f"sigma_{m}"]),
                        )
                    else:
                        y_dist = dist.Bernoulli(logits=y_loc)

                    feature_idx = slice(
                        self.feature_offsets[m],
                        self.feature_offsets[m + 1],
                    )

                    with pyro.poutine.mask(
                        mask=mask[..., feature_idx]
                    ), pyro.poutine.scale(scale=self.view_scales[m]):
                        output_dict[f"y_{m}"] = pyro.sample(
                            f"y_{m}",
                            y_dist,
                            obs=obs[..., feature_idx],
                            infer={"is_auxiliary": True},
                        )

        return output_dict


class MuVIGuide(PyroModule):
    def __init__(
        self,
        model,
        init_loc: float = 0.0,
        init_scale: float = 0.1,
    ):
        """Approximate posterior.

        Parameters
        ----------
        model : object
            A MuVIModel object
        init_loc : float, optional
            Initial value for loc parameters, by default 0.0
        init_scale : float, optional
            Initial value for scale parameters, by default 0.1
        """
        super().__init__(name="MuVIGuide")
        self.model = model
        self.locs = PyroModule()
        self.scales = PyroModule()

        self.init_loc = init_loc
        self.init_scale = init_scale

        self.site_to_dist = self.setup()

    def _get_loc_and_scale(self, name: str):
        """Get loc and scale parameters.

        Parameters
        ----------
        name : str
            Name of the sampling site

        Returns
        -------
        tuple
            Tuple of (loc, scale)
        """
        site_loc = deep_getattr(self.locs, name)
        site_scale = deep_getattr(self.scales, name)
        return site_loc, site_scale

    def setup(self):
        """Setup parameters and sampling sites."""
        n_views = self.model.n_views
        n_samples = self.model.n_samples
        n_factors = self.model.n_factors
        n_features = self.model.n_features
        n_covariates = self.model.n_covariates

        site_to_shape = {
            "z": (n_samples, n_factors),
            "view_scale": (n_views,),
            "factor_scale": (n_factors, n_views),
        }

        normal_sites = ["z"]
        for m in range(n_views):
            site_to_shape[f"local_scale_{m}"] = (n_factors, n_features[m])
            site_to_shape[f"caux_{m}"] = (n_factors, n_features[m])
            site_to_shape[f"w_{m}"] = (n_factors, n_features[m])
            site_to_shape[f"beta_{m}"] = (n_covariates, n_features[m])
            site_to_shape[f"sigma_{m}"] = (n_features[m],)

            normal_sites.append(f"w_{m}")
            normal_sites.append(f"beta_{m}")

        site_to_dist = {
            k: "Normal" if k in normal_sites else "LogNormal" for k in site_to_shape
        }

        for name, shape in site_to_shape.items():
            deep_setattr(
                self.locs,
                name,
                PyroParam(
                    self.init_loc * self.model._ones(shape),
                    constraints.real,
                ),
            )

            deep_setattr(
                self.scales,
                name,
                PyroParam(
                    self.init_scale * self.model._ones(shape),
                    constraints.softplus_positive,
                ),
            )

        return site_to_dist

    @torch.no_grad()
    def mode(self, name: str):
        """Get the MAP estimates.

        Parameters
        ----------
        name : str
            Name of the sampling site

        Returns
        -------
        torch.Tensor
            MAP estimate
        """
        loc, scale = self._get_loc_and_scale(name)
        mode = loc
        if self.site_to_dist[name] == "LogNormal":
            mode = (loc - scale.square()).exp()
        if (name == "z") and any(self.model.nmf):
            mode = self.model.pos_transform(mode)
        for m in range(self.model.n_views):
            if name == f"w_{m}" and self.model.nmf[m]:
                mode = self.model.pos_transform(mode)
        return mode.clone()

    @torch.no_grad()
    def _get_map_estimate(self, param_name: str, is_list_param: bool, as_list: bool):
        is_list_param |= as_list
        if is_list_param:
            params = [self.mode(f"{param_name}_{m}") for m in range(self.model.n_views)]
        else:
            params = [self.mode(param_name)]

        vector_params = ["view_scale", "sigma"]

        if param_name in vector_params:
            params = [param[None, :] for param in params]
        params = [param.cpu().detach().numpy() for param in params]

        if as_list:
            return params
        return np.concatenate(params, axis=1)

    def get_sigma(self, as_list: bool = False):
        """Get the marginal feature scales."""
        return self._get_map_estimate("sigma", True, as_list=as_list)

    def get_view_scale(self):
        """Get the view scales."""
        return self._get_map_estimate("view_scale", False, False)

    def get_factor_scale(self):
        """Get the factor scales."""
        return self._get_map_estimate("factor_scale", False, False).T

    def get_local_scale(self, as_list: bool = False):
        """Get the local scales."""
        return self._get_map_estimate("local_scale", True, as_list=as_list)

    def get_caux(self, as_list: bool = False):
        """Get the c auxiliaries."""
        return self._get_map_estimate("caux", True, as_list=as_list)

    def get_z(self):
        """Get the factor scores."""
        return self._get_map_estimate("z", False, False)

    def get_w(self, as_list: bool = False):
        """Get the factor loadings."""
        return self._get_map_estimate("w", True, as_list=as_list)

    def get_beta(self, as_list: bool = False):
        """Get the beta coefficients."""
        return self._get_map_estimate("beta", True, as_list=as_list)

    def _sample(self, name, index=None):
        loc, scale = self._get_loc_and_scale(name)
        if index is not None:
            loc = loc.index_select(0, index)
            scale = scale.index_select(0, index)
        if self.site_to_dist[name] == "LogNormal":
            return pyro.sample(name, dist.LogNormal(loc, scale))
        if self.site_to_dist[name] == "Dirichlet":
            return pyro.sample(name, dist.Dirichlet(scale))
        return pyro.sample(name, dist.Normal(loc, scale))

    def forward(
        self,
        sample_idx: torch.Tensor,
        obs: torch.Tensor,
        mask: torch.Tensor,
        covs: Optional[torch.Tensor] = None,
        prior_scales: Optional[torch.Tensor] = None,
    ):
        """Approximate posterior."""
        output_dict = {}

        view_plate = self.model.get_plate("view")
        factor_l_plate = self.model.get_plate("factor_left")
        feature_plates = [
            self.model.get_plate(f"feature_{m}") for m in range(self.model.n_views)
        ]
        sample_plate = self.model.get_plate("sample", subsample=sample_idx)

        if self.model.n_covariates > 0:
            covariate_plate = self.model.get_plate("covariate")

        with view_plate:
            output_dict["view_scale"] = self._sample("view_scale")
            with factor_l_plate:
                output_dict["factor_scale"] = self._sample("factor_scale")

        for m in range(self.model.n_views):
            with feature_plates[m]:
                with factor_l_plate:
                    output_dict[f"local_scale_{m}"] = self._sample(f"local_scale_{m}")
                    if self.model.reg_hs:
                        output_dict[f"caux_{m}"] = self._sample(f"caux_{m}")
                    output_dict[f"w_{m}"] = self._sample(f"w_{m}")

                if self.model.n_covariates > 0:
                    with covariate_plate:
                        output_dict[f"beta_{m}"] = self._sample(f"beta_{m}")

                output_dict[f"sigma_{m}"] = self._sample(f"sigma_{m}")

        with sample_plate as indices:
            output_dict["z"] = self._sample("z", indices)
        return output_dict



"""Data module."""
import itertools
import logging
import math

from typing import List
from typing import Optional
from typing import Tuple

import anndata as ad
import mudata as mu
import numpy as np


logger = logging.getLogger(__name__)


class DataGenerator:
    def __init__(
        self,
        n_samples: int = 1000,
        n_features: Optional[List[int]] = None,
        likelihoods: Optional[List[str]] = None,
        n_fully_shared_factors: int = 2,
        n_partially_shared_factors: int = 15,
        n_private_factors: int = 3,
        factor_size_params: Optional[Tuple[float]] = None,
        factor_size_dist: str = "uniform",
        n_active_factors: float = 1.0,
        n_covariates: int = 0,
        n_response: int = 0,
        nmf: Optional[List[bool]] = None,
        **kwargs,
    ) -> None:
        """Generate synthetic data

        Parameters
        ----------
        n_samples : int, optional
            Number of samples, by default 1000
        n_features : List[int], optional
            Number of features for each view, by default None
        likelihoods : List[str], optional
            Likelihoods for each view, 'normal' or 'bernoulli', by default None
        n_fully_shared_factors : int, optional
            Number of fully shared latent factors, by default 2
        n_partially_shared_factors : int, optional
            Number of partially shared latent factors, by default 15
        n_private_factors : int, optional
            Number of private latent factors, by default 3
        factor_size_params : Tuple[float], optional
            Parameters for the distribution of the number
            of active factor loadings for the latent factors,
            by default None
        factor_size_dist : str, optional
            Distribution of the number of active factor loadings,
            either "uniform" or "gamma",
            by default "uniform"
        n_active_factors : float, optional
            Number or fraction of active factors, by default 1.0 (all)
        n_covariates : int, optional
            Number of observed covariates, by default 0
        n_response : int, optional
            Number of response variables from the latent factors, by default 0
        nmf : List[bool], optional
            Whether to generate data from a non-negative matrix factorization,
            by default False
        """

        self.n_samples = n_samples
        self.n_features = n_features
        self.n_views = len(self.n_features)
        self.n_fully_shared_factors = n_fully_shared_factors
        self.n_partially_shared_factors = n_partially_shared_factors
        self.n_private_factors = n_private_factors
        self.n_covariates = n_covariates
        self.n_response = n_response

        if factor_size_params is None:
            if factor_size_dist == "uniform":
                logger.warning(
                    "Using a uniform distribution with parameters 0.05 and 0.15 "
                    "for generating the number of active factor loadings."
                )
                factor_size_params = (0.05, 0.15)
            elif factor_size_dist == "gamma":
                logger.warning(
                    "Using a uniform distribution with shape of 1 and scale of 50 "
                    "for generating the number of active factor loadings."
                )
                factor_size_params = (1.0, 50.0)

        if isinstance(factor_size_params, tuple):
            factor_size_params = [factor_size_params for _ in range(self.n_views)]

        self.factor_size_params = factor_size_params
        self.factor_size_dist = factor_size_dist

        # custom assignment
        if likelihoods is None:
            likelihoods = ["normal" for _ in range(self.n_views)]
        self.likelihoods = likelihoods

        self.n_active_factors = n_active_factors

        if nmf is None:
            nmf = [False for _ in range(self.n_views)]
        self.nmf = nmf

        # set upon data generation
        # covariates
        self.x = None
        # covariate coefficients
        self.betas = None
        # latent factors
        self.z = None
        # factor loadings
        self.ws = None
        self.sigmas = None
        self.ys = None
        self.w_masks = None
        self.noisy_w_masks = None
        self.active_factor_indices = None
        self.view_factor_mask = None
        # set when introducing missingness
        self.presence_masks = None

        self.response_w = None
        self.response_sigma = None
        self.response = None

    @property
    def n_factors(self):
        return (
            self.n_fully_shared_factors
            + self.n_partially_shared_factors
            + self.n_private_factors
        )

    def _to_matrix(self, matrix_list):
        return np.concatenate(matrix_list, axis=1)

    def _attr_to_matrix(self, attr_name):
        attr = getattr(self, attr_name)
        if attr is not None:
            attr = self._to_matrix(attr)
        return attr

    def _mask_to_nan(self):
        nan_masks = []
        for mask in self.presence_masks:
            nan_mask = np.array(mask, dtype=np.float32, copy=True)
            nan_mask[nan_mask == 0] = np.nan
            nan_masks.append(nan_mask)
        return nan_masks

    def _mask_to_bool(self):
        bool_masks = []
        for mask in self.presence_masks:
            bool_mask = mask == 1.0
            bool_masks.append(bool_mask)
        return bool_masks

    @property
    def missing_ys(self):
        if self.ys is None:
            logger.warning("Generate data first by calling `generate`.")
            return []
        if self.presence_masks is None:
            logger.warning(
                "Introduce missing data first by calling `generate_missingness`."
            )
            return self.ys

        nan_masks = self._mask_to_nan()

        return [self.ys[m] * nan_masks[m] for m in range(self.n_views)]

    @property
    def y(self):
        return self._attr_to_matrix("ys")

    @property
    def missing_y(self):
        return self._attr_to_matrix("missing_ys")

    @property
    def w(self):
        return self._attr_to_matrix("ws")

    @property
    def w_mask(self):
        return self._attr_to_matrix("w_masks")

    @property
    def noisy_w_mask(self):
        return self._attr_to_matrix("noisy_w_masks")

    def _generate_view_factor_mask(self, rng=None, all_combs=False):
        if all_combs and self.n_views == 1:
            logger.warning(
                "Single view dataset, "
                "cannot generate factor combinations for a single view."
            )
            all_combs = False
        if all_combs:
            logger.warning(
                "Generating all possible binary combinations of "
                f"{self.n_views} variables."
            )
            self.n_fully_shared_factors = 1
            self.n_private_factors = self.n_views
            self.n_partially_shared_factors = (
                2**self.n_views - 2 - self.n_private_factors
            )
            logger.warning(
                "New factor configuration: "
                f"{self.n_fully_shared_factors} fully shared, "
                f"{self.n_partially_shared_factors} partially shared, "
                f"{self.n_private_factors} private factors."
            )

            return np.array(
                [list(i) for i in itertools.product([1, 0], repeat=self.n_views)]
            )[:-1, :].T
        if rng is None:
            rng = np.random.default_rng()

        view_factor_mask = np.ones([self.n_views, self.n_factors])

        for factor_idx in range(self.n_fully_shared_factors, self.n_factors):
            # exclude view subsets for partially shared factors
            if (
                factor_idx
                < self.n_fully_shared_factors + self.n_partially_shared_factors
            ):
                if self.n_views > 2:
                    exclude_view_subset_size = rng.integers(1, self.n_views - 1)
                else:
                    exclude_view_subset_size = 0

                exclude_view_subset = rng.choice(
                    self.n_views, exclude_view_subset_size, replace=False
                )
            # exclude all but one view for private factors
            else:
                include_view_idx = rng.integers(self.n_views)
                exclude_view_subset = [
                    i for i in range(self.n_views) if i != include_view_idx
                ]

            for m in exclude_view_subset:
                view_factor_mask[m, factor_idx] = 0

        if self.n_private_factors >= self.n_views:
            view_factor_mask[-self.n_views :, -self.n_views :] = np.eye(self.n_views)

        return view_factor_mask

    def normalise(self, with_std=False):
        for m in range(self.n_views):
            if self.likelihoods[m] == "normal":
                y = np.array(self.ys[m], dtype=np.float32, copy=True)
                y -= y.mean(axis=0)
                if with_std:
                    y_std = y.std(axis=0)
                    y = np.divide(y, y_std, out=np.zeros_like(y), where=y_std != 0)
                self.ys[m] = y

    def sigmoid(self, x):
        return 1.0 / (1 + np.exp(-x))

    def generate(
        self,
        seed: Optional[int] = None,
        all_combs: bool = False,
        overwrite: bool = False,
    ) -> None:
        rng = np.random.default_rng()

        if seed is not None:
            rng = np.random.default_rng(seed)

        if self.ys is not None and not overwrite:
            logger.warning(
                "Data has already been generated, "
                "to generate new data please set `overwrite` to True."
            )
            return rng

        view_factor_mask = self._generate_view_factor_mask(rng, all_combs)

        n_active_factors = self.n_active_factors
        if n_active_factors <= 1.0:
            # if fraction of active factors convert to int
            n_active_factors = int(n_active_factors * self.n_factors)

        active_factor_indices = sorted(
            rng.choice(
                self.n_factors,
                size=math.ceil(n_active_factors),
                replace=False,
            )
        )

        # generate factor scores which lie in the latent space
        z = rng.standard_normal((self.n_samples, self.n_factors))

        if any(self.nmf):
            z = np.abs(z)

        if self.n_covariates > 0:
            x = rng.standard_normal((self.n_samples, self.n_covariates))
            if any(self.nmf):
                x = np.abs(x)

        betas = []
        ws = []
        sigmas = []
        ys = []
        w_masks = []

        for factor_idx in range(self.n_factors):
            if factor_idx not in active_factor_indices:
                view_factor_mask[:, factor_idx] = 0.0

        for m in range(self.n_views):
            n_features = self.n_features[m]
            w_shape = (self.n_factors, n_features)
            w = rng.standard_normal(w_shape)
            w_mask = np.zeros(w_shape)

            fraction_active_features = {
                "gamma": (
                    lambda shape, scale, n_features=n_features: (
                        rng.gamma(shape, scale, self.n_factors) + 20
                    )
                    / n_features
                ),
                "uniform": lambda low, high, n_features=n_features: rng.uniform(
                    low, high, self.n_factors
                ),
            }[self.factor_size_dist](
                self.factor_size_params[m][0], self.factor_size_params[m][1]
            )

            for factor_idx, faft in enumerate(fraction_active_features):
                if view_factor_mask[m, factor_idx] > 0:
                    w_mask[factor_idx] = rng.choice(2, n_features, p=[1 - faft, faft])

            # set small values to zero
            tiny_w_threshold = 0.1
            w_mask[np.abs(w) < tiny_w_threshold] = 0.0
            w_mask = w_mask.astype(bool)
            # add some noise to avoid exactly zero values
            w = np.where(w_mask, w, rng.standard_normal(w_shape) / 100)
            assert ((np.abs(w) > tiny_w_threshold) == w_mask).all()

            if self.nmf[m]:
                w = np.abs(w)

            y_loc = np.matmul(z, w)

            if self.n_covariates > 0:
                beta_shape = (self.n_covariates, n_features)
                # reduce effect of betas by scaling them down
                beta = rng.standard_normal(beta_shape) / 10
                if self.nmf[m]:
                    beta = np.abs(beta)
                y_loc = y_loc + np.matmul(x, beta)
                betas.append(beta)

            # generate feature sigmas
            sigma = 1.0 / np.sqrt(rng.gamma(10.0, 1.0, n_features))

            if self.likelihoods[m] == "normal":
                y = rng.normal(loc=y_loc, scale=sigma)
                if self.nmf[m]:
                    y = np.abs(y)
            elif self.likelihoods[m] == "bernoulli":
                y = rng.binomial(1, self.sigmoid(y_loc))
            elif self.likelihoods[m] == "poisson":
                rate = np.exp(y_loc)
                y = rng.poisson(rate)

            ws.append(w)
            sigmas.append(sigma)
            ys.append(y)
            w_masks.append(w_mask)

        if self.n_covariates > 0:
            self.x = x
            self.betas = betas

        self.z = z
        self.ws = ws
        self.w_masks = w_masks
        self.sigmas = sigmas
        self.ys = ys
        self.active_factor_indices = active_factor_indices
        self.view_factor_mask = view_factor_mask

        if self.n_response > 0:
            self.response_w = rng.standard_normal((self.n_factors, self.n_response))
            self.response_sigma = 1.0 / np.sqrt(rng.gamma(10.0, 1.0, self.n_response))
            self.response = rng.normal(
                loc=np.matmul(z, self.response_w), scale=self.response_sigma
            )

        return rng

    def get_noisy_mask(self, rng=None, noise_fraction=0.1, informed_view_indices=None):
        if rng is None:
            rng = np.random.default_rng()

        if informed_view_indices is None:
            logger.warning(
                "Parameter `informed_view_indices` set to None, "
                "adding noise to all views."
            )
            informed_view_indices = list(range(self.n_views))

        noisy_w_masks = [np.array(mask, copy=True) for mask in self.w_masks]

        if len(informed_view_indices) == 0:
            logger.warning(
                "Parameter `informed_view_indices` "
                "set to an empty list, removing information from all views."
            )
            self.noisy_w_masks = [np.ones_like(mask) for mask in noisy_w_masks]
            return self.noisy_w_masks

        for m in range(self.n_views):
            noisy_w_mask = noisy_w_masks[m]

            if m in informed_view_indices:
                fraction_active_cells = (
                    noisy_w_mask.mean(axis=1).sum() / self.view_factor_mask[0].sum()
                )
                for factor_idx in range(self.n_factors):
                    active_cell_indices = noisy_w_mask[factor_idx, :].nonzero()[0]
                    # if all features turned off
                    # => simulate random noise in terms of false positives only
                    if len(active_cell_indices) == 0:
                        logger.warning(
                            f"Factor {factor_idx} is completely off, inserting "
                            f"{(100 * fraction_active_cells):.2f}%% false positives.",
                        )
                        active_cell_indices = rng.choice(
                            self.n_features[m],
                            int(self.n_features[m] * fraction_active_cells),
                            replace=False,
                        )

                    inactive_cell_indices = (
                        noisy_w_mask[factor_idx, :] == 0
                    ).nonzero()[0]
                    n_noisy_cells = int(noise_fraction * len(active_cell_indices))
                    swapped_indices = zip(
                        rng.choice(
                            len(active_cell_indices), n_noisy_cells, replace=False
                        ),
                        rng.choice(
                            len(inactive_cell_indices), n_noisy_cells, replace=False
                        ),
                    )

                    for on_idx, off_idx in swapped_indices:
                        noisy_w_mask[factor_idx, active_cell_indices[on_idx]] = 0.0
                        noisy_w_mask[factor_idx, inactive_cell_indices[off_idx]] = 1.0

            else:
                noisy_w_mask.fill(0.0)

        self.noisy_w_masks = noisy_w_masks
        return self.noisy_w_masks

    def generate_missingness(
        self,
        random_fraction: float = 0.0,
        n_partial_samples: int = 0,
        n_partial_features: int = 0,
        missing_fraction_partial_features: float = 0.0,
        seed=None,
    ):
        rng = np.random.default_rng()

        if seed is not None:
            rng = np.random.default_rng(seed)

        n_partial_samples = int(n_partial_samples)
        n_partial_features = int(n_partial_features)

        sample_view_mask = np.ones((self.n_samples, self.n_views))
        missing_sample_indices = rng.choice(
            self.n_samples, n_partial_samples, replace=False
        )

        # partially missing samples
        for ms_idx in missing_sample_indices:
            if self.n_views > 1:
                exclude_view_subset_size = rng.integers(1, self.n_views)
            else:
                exclude_view_subset_size = 0
            exclude_view_subset = rng.choice(
                self.n_views, exclude_view_subset_size, replace=False
            )
            sample_view_mask[ms_idx, exclude_view_subset] = 0

        mask = np.repeat(sample_view_mask, self.n_features, axis=1)

        # partially missing features
        missing_feature_indices = rng.choice(
            sum(self.n_features), n_partial_features, replace=False
        )

        for mf_idx in missing_feature_indices:
            random_sample_indices = rng.choice(
                self.n_samples,
                int(self.n_samples * missing_fraction_partial_features),
                replace=False,
            )
            mask[random_sample_indices, mf_idx] = 0

        # remove random fraction
        mask *= rng.choice([0, 1], mask.shape, p=[random_fraction, 1 - random_fraction])

        view_feature_offsets = [0, *np.cumsum(self.n_features).tolist()]
        masks = []
        for offset_idx in range(len(view_feature_offsets) - 1):
            start_offset = view_feature_offsets[offset_idx]
            end_offset = view_feature_offsets[offset_idx + 1]
            masks.append(mask[:, start_offset:end_offset])

        self.presence_masks = masks

        return rng

    def _permute_features(self, lst, new_order):
        return [np.array(lst[m][:, o], copy=True) for m, o in enumerate(new_order)]

    def _permute_factors(self, lst, new_order):
        return [np.array(lst[m][o, :], copy=True) for m, o in enumerate(new_order)]

    def permute_features(self, new_feature_order):
        if len(new_feature_order) != self.n_features:
            raise ValueError(
                "Length of new order list must equal the number of features."
            )

        if self.betas is not None:
            self.betas = self._permute_features(self.betas, new_feature_order)
        self.ws = self._permute_features(self.ws, new_feature_order)
        self.w_masks = self._permute_features(self.w_masks, new_feature_order)
        if self.noisy_w_masks is not None:
            self.noisy_w_masks = self._permute_features(
                self.noisy_w_masks, new_feature_order
            )
        self.sigmas = self._permute_features(self.sigmas, new_feature_order)
        self.ys = self._permute_features(self.ys, new_feature_order)
        if self.presence_masks is not None:
            self.missing_ys = self._permute_features(self.missing_ys, new_feature_order)
            self.presence_masks = self._permute_features(
                self.presence_masks, new_feature_order
            )

    def permute_factors(self, new_factor_order):
        if len(new_factor_order) != self.n_factors:
            raise ValueError(
                "Length of new order list must equal the number of factors."
            )

        self.z = np.array(self.z[:, np.array(new_factor_order)], copy=True)
        self.ws = self._permute_factors(self.ws, new_factor_order)
        self.w_masks = self._permute_factors(self.w_masks, new_factor_order)
        if self.noisy_w_masks is not None:
            self.noisy_w_masks = self._permute_factors(
                self.noisy_w_masks, new_factor_order
            )
        self.view_factor_mask = [
            self.view_factor_mask[m, np.array(new_factor_order)]
            for m in range(self.n_views)
        ]
        self.active_factor_indices = np.array(
            self.active_factor_indices[np.array(new_factor_order)], copy=True
        )

    def to_mdata(self) -> mu.MuData:
        view_names = []
        ad_dict = {}
        for m in range(self.n_views):
            adata = ad.AnnData(
                self.ys[m],
                dtype=np.float32,
            )
            adata.var_names = f"feature_group_{m}:" + adata.var_names
            adata.varm["w"] = self.ws[m].T
            adata.varm["w_mask"] = self.w_masks[m].T
            view_name = f"feature_group_{m}"
            ad_dict[view_name] = adata
            view_names.append(view_name)

        mdata = mu.MuData(ad_dict)
        mdata.uns["likelihoods"] = dict(zip(view_names, self.likelihoods))
        mdata.uns["n_active_factors"] = self.n_active_factors
        mdata.obsm["x"] = self.x
        mdata.obsm["z"] = self.z

        return mdata



import logging

import anndata as ad
import numpy as np
import pandas as pd


logger = logging.getLogger(__name__)


class Cache:
    FILTERED_KEY = "filtered_scores"
    META_KEY = "metadata"

    METRIC_RMSE = "rmse"
    METRIC_R2 = "r2"

    TEST_POS = "pos"
    TEST_NEG = "neg"
    TEST_ALL = "all"
    TEST_P = "p"
    TEST_P_ADJ = "p_adj"
    TEST_T = "t"

    UNS_GROUPED_R2 = "grouped_r2"

    def __init__(self, model) -> None:
        self.clear()
        self.setup(model)

    def clear(self):
        self.uns = {}
        self.factor_adata = None
        self.cov_adata = None
        self.use_rep = "X"

    @property
    def factor_metadata(self):
        if self.factor_adata is None:
            return None
        return self.factor_adata.varm[Cache.META_KEY]

    @property
    def cov_metadata(self):
        if self.cov_adata is None:
            return None
        return self.cov_adata.varm[Cache.META_KEY]

    def setup(self, model):
        self._setup_uns()
        self._setup_adata(model)

    def _setup_uns(self):
        for view_score_key in [Cache.METRIC_R2, Cache.METRIC_RMSE]:
            if view_score_key not in self.uns:
                self.uns[f"view_{view_score_key}"] = {}

    def _setup_adata(self, model):
        columns = []
        for key in [Cache.METRIC_R2, Cache.METRIC_RMSE]:
            for vn in model.view_names:
                columns.append(f"{key}_{vn}")

        for key in [Cache.TEST_P, Cache.TEST_P_ADJ, Cache.TEST_T]:
            for sign in [Cache.TEST_ALL, Cache.TEST_NEG, Cache.TEST_POS]:
                for vn in model.view_names:
                    columns.append(f"{key}_{sign}_{vn}")

        if model.n_factors > 0:
            self.factor_adata = ad.AnnData(
                model.get_factor_scores(as_df=True), dtype=np.float32
            )
            self.factor_adata.varm[Cache.META_KEY] = pd.DataFrame(
                index=self.factor_adata.var_names, columns=columns, dtype=np.float32
            )

        if model.n_covariates > 0:
            self.cov_adata = ad.AnnData(
                model.get_covariates(as_df=True), dtype=np.float32
            )
            self.cov_adata.varm[Cache.META_KEY] = pd.DataFrame(
                index=self.cov_adata.var_names, columns=columns, dtype=np.float32
            )

    def update_uns(self, key, scores):
        self.uns[key].update(scores)

    def update_factor_metadata(self, scores):
        if self.factor_adata is not None:
            self.factor_adata.varm[Cache.META_KEY].update(scores)

    def update_cov_metadata(self, scores):
        if self.cov_adata is not None:
            self.cov_adata.varm[Cache.META_KEY].update(scores)

    def filter_factors(self, factor_idx):
        self.factor_adata.obsm[Cache.FILTERED_KEY] = (
            self.factor_adata.to_df().loc[:, factor_idx].copy()
        )
        self.use_rep = Cache.FILTERED_KEY
        uns_keys = list(self.factor_adata.uns.keys())

        # remove neighborhood information
        if "neighbors" in uns_keys:
            logger.warning("Removing old neighborhood graph.")
            self.factor_adata.uns.pop("neighbors", None)
            logger.warning("Removing old distances.")
            self.factor_adata.obsp.pop("distances", None)
            logger.warning("Removing old connectivities.")
            self.factor_adata.obsp.pop("connectivities", None)

        # remove dendrogram information
        for key in uns_keys:
            if "dendrogram" in key:
                logger.warning("Removing old dendrogram.")
                self.factor_adata.uns.pop(key, None)

        logger.info("Factors filtered successfully.")

        return True



import torch


def get_free_gpu_idx():
    """Get the index of the GPU with current lowest memory usage."""

    max_free_idx = 0
    max_free_mem = torch.cuda.mem_get_info(0)[0]
    for i in range(torch.cuda.device_count()):
        if torch.cuda.mem_get_info(i)[0] > max_free_mem:
            max_free_idx = i
            max_free_mem = torch.cuda.mem_get_info(i)[0]

    return max_free_idx
